# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

__doc__ = """
    Définit les objets numériques génériques.
"""
__author__ = "Jean-Philippe ARGAUD"

import os, copy, types, sys, logging, numpy
from daCore.BasicObjects import Operator, Covariance, PartialAlgorithm
from daCore.PlatformInfo import PlatformInfo
mpr = PlatformInfo().MachinePrecision()
mfp = PlatformInfo().MaximumPrecision()
# logging.getLogger().setLevel(logging.DEBUG)

# ==============================================================================
def ExecuteFunction( triplet ):
    assert len(triplet) == 3, "Incorrect number of arguments"
    X, xArgs, funcrepr = triplet
    __X = numpy.ravel( X ).reshape((-1,1))
    __sys_path_tmp = sys.path ; sys.path.insert(0,funcrepr["__userFunction__path"])
    __module = __import__(funcrepr["__userFunction__modl"], globals(), locals(), [])
    __fonction = getattr(__module,funcrepr["__userFunction__name"])
    sys.path = __sys_path_tmp ; del __sys_path_tmp
    if isinstance(xArgs, dict):
        __HX  = __fonction( __X, **xArgs )
    else:
        __HX  = __fonction( __X )
    return numpy.ravel( __HX )

# ==============================================================================
class FDApproximation(object):
    """
    Cette classe sert d'interface pour définir les opérateurs approximés. A la
    création d'un objet, en fournissant une fonction "Function", on obtient un
    objet qui dispose de 3 méthodes "DirectOperator", "TangentOperator" et
    "AdjointOperator". On contrôle l'approximation DF avec l'incrément
    multiplicatif "increment" valant par défaut 1%, ou avec l'incrément fixe
    "dX" qui sera multiplié par "increment" (donc en %), et on effectue de DF
    centrées si le booléen "centeredDF" est vrai.
    """
    def __init__(self,
            name                  = "FDApproximation",
            Function              = None,
            centeredDF            = False,
            increment             = 0.01,
            dX                    = None,
            extraArguments        = None,
            reducingMemoryUse     = False,
            avoidingRedundancy    = True,
            toleranceInRedundancy = 1.e-18,
            lenghtOfRedundancy    = -1,
            mpEnabled             = False,
            mpWorkers             = None,
            mfEnabled             = False,
            ):
        self.__name = str(name)
        self.__extraArgs = extraArguments
        #
        if mpEnabled:
            try:
                import multiprocessing
                self.__mpEnabled = True
            except ImportError:
                self.__mpEnabled = False
        else:
            self.__mpEnabled = False
        self.__mpWorkers = mpWorkers
        if self.__mpWorkers is not None and self.__mpWorkers < 1:
            self.__mpWorkers = None
        logging.debug("FDA Calculs en multiprocessing : %s (nombre de processus : %s)"%(self.__mpEnabled,self.__mpWorkers))
        #
        self.__mfEnabled = bool(mfEnabled)
        logging.debug("FDA Calculs en multifonctions : %s"%(self.__mfEnabled,))
        #
        self.__rmEnabled = bool(reducingMemoryUse)
        logging.debug("FDA Calculs avec réduction mémoire : %s"%(self.__rmEnabled,))
        #
        if avoidingRedundancy:
            self.__avoidRC = True
            self.__tolerBP = float(toleranceInRedundancy)
            self.__lenghtRJ = int(lenghtOfRedundancy)
            self.__listJPCP = [] # Jacobian Previous Calculated Points
            self.__listJPCI = [] # Jacobian Previous Calculated Increment
            self.__listJPCR = [] # Jacobian Previous Calculated Results
            self.__listJPPN = [] # Jacobian Previous Calculated Point Norms
            self.__listJPIN = [] # Jacobian Previous Calculated Increment Norms
        else:
            self.__avoidRC = False
        logging.debug("FDA Calculs avec réduction des doublons : %s"%self.__avoidRC)
        if self.__avoidRC:
            logging.debug("FDA Tolérance de détermination des doublons : %.2e"%self.__tolerBP)
        #
        if self.__mpEnabled:
            if isinstance(Function,types.FunctionType):
                logging.debug("FDA Calculs en multiprocessing : FunctionType")
                self.__userFunction__name = Function.__name__
                try:
                    mod = os.path.join(Function.__globals__['filepath'],Function.__globals__['filename'])
                except Exception:
                    mod = os.path.abspath(Function.__globals__['__file__'])
                if not os.path.isfile(mod):
                    raise ImportError("No user defined function or method found with the name %s"%(mod,))
                self.__userFunction__modl = os.path.basename(mod).replace('.pyc','').replace('.pyo','').replace('.py','')
                self.__userFunction__path = os.path.dirname(mod)
                del mod
                self.__userOperator = Operator(
                    name                 = self.__name,
                    fromMethod           = Function,
                    avoidingRedundancy   = self.__avoidRC,
                    inputAsMultiFunction = self.__mfEnabled,
                    extraArguments       = self.__extraArgs )
                self.__userFunction = self.__userOperator.appliedTo # Pour le calcul Direct
            elif isinstance(Function,types.MethodType):
                logging.debug("FDA Calculs en multiprocessing : MethodType")
                self.__userFunction__name = Function.__name__
                try:
                    mod = os.path.join(Function.__globals__['filepath'],Function.__globals__['filename'])
                except Exception:
                    mod = os.path.abspath(Function.__func__.__globals__['__file__'])
                if not os.path.isfile(mod):
                    raise ImportError("No user defined function or method found with the name %s"%(mod,))
                self.__userFunction__modl = os.path.basename(mod).replace('.pyc','').replace('.pyo','').replace('.py','')
                self.__userFunction__path = os.path.dirname(mod)
                del mod
                self.__userOperator = Operator(
                    name                 = self.__name,
                    fromMethod           = Function,
                    avoidingRedundancy   = self.__avoidRC,
                    inputAsMultiFunction = self.__mfEnabled,
                    extraArguments       = self.__extraArgs )
                self.__userFunction = self.__userOperator.appliedTo # Pour le calcul Direct
            else:
                raise TypeError("User defined function or method has to be provided for finite differences approximation.")
        else:
            self.__userOperator = Operator(
                name                 = self.__name,
                fromMethod           = Function,
                avoidingRedundancy   = self.__avoidRC,
                inputAsMultiFunction = self.__mfEnabled,
                extraArguments       = self.__extraArgs )
            self.__userFunction = self.__userOperator.appliedTo
        #
        self.__centeredDF = bool(centeredDF)
        if abs(float(increment)) > 1.e-15:
            self.__increment  = float(increment)
        else:
            self.__increment  = 0.01
        if dX is None:
            self.__dX     = None
        else:
            self.__dX     = numpy.ravel( dX )

    # ---------------------------------------------------------
    def __doublon__(self, __e, __l, __n, __v=None):
        __ac, __iac = False, -1
        for i in range(len(__l)-1,-1,-1):
            if numpy.linalg.norm(__e - __l[i]) < self.__tolerBP * __n[i]:
                __ac, __iac = True, i
                if __v is not None: logging.debug("FDA Cas%s déjà calculé, récupération du doublon %i"%(__v,__iac))
                break
        return __ac, __iac

    # ---------------------------------------------------------
    def __listdotwith__(self, __LMatrix, __dotWith = None, __dotTWith = None):
        "Produit incrémental d'une matrice liste de colonnes avec un vecteur"
        if not isinstance(__LMatrix, (list,tuple)):
            raise TypeError("Columnwise list matrix has not the proper type: %s"%type(__LMatrix))
        if __dotWith is not None:
            __Idwx = numpy.ravel( __dotWith )
            assert len(__LMatrix) == __Idwx.size, "Incorrect size of elements"
            __Produit = numpy.zeros(__LMatrix[0].size)
            for i, col in enumerate(__LMatrix):
                __Produit += float(__Idwx[i]) * col
            return __Produit
        elif __dotTWith is not None:
            _Idwy = numpy.ravel( __dotTWith ).T
            assert __LMatrix[0].size == _Idwy.size, "Incorrect size of elements"
            __Produit = numpy.zeros(len(__LMatrix))
            for i, col in enumerate(__LMatrix):
                __Produit[i] = float( _Idwy @ col)
            return __Produit
        else:
            __Produit = None
        return __Produit

    # ---------------------------------------------------------
    def DirectOperator(self, X, **extraArgs ):
        """
        Calcul du direct à l'aide de la fonction fournie.

        NB : les extraArgs sont là pour assurer la compatibilité d'appel, mais
        ne doivent pas être données ici à la fonction utilisateur.
        """
        logging.debug("FDA Calcul DirectOperator (explicite)")
        if self.__mfEnabled:
            _HX = self.__userFunction( X, argsAsSerie = True )
        else:
            _HX = numpy.ravel(self.__userFunction( numpy.ravel(X) ))
        #
        return _HX

    # ---------------------------------------------------------
    def TangentMatrix(self, X, dotWith = None, dotTWith = None ):
        """
        Calcul de l'opérateur tangent comme la Jacobienne par différences finies,
        c'est-à-dire le gradient de H en X. On utilise des différences finies
        directionnelles autour du point X. X est un numpy.ndarray.

        Différences finies centrées (approximation d'ordre 2):
        1/ Pour chaque composante i de X, on ajoute et on enlève la perturbation
           dX[i] à la  composante X[i], pour composer X_plus_dXi et X_moins_dXi, et
           on calcule les réponses HX_plus_dXi = H( X_plus_dXi ) et HX_moins_dXi =
           H( X_moins_dXi )
        2/ On effectue les différences (HX_plus_dXi-HX_moins_dXi) et on divise par
           le pas 2*dXi
        3/ Chaque résultat, par composante, devient une colonne de la Jacobienne

        Différences finies non centrées (approximation d'ordre 1):
        1/ Pour chaque composante i de X, on ajoute la perturbation dX[i] à la
           composante X[i] pour composer X_plus_dXi, et on calcule la réponse
           HX_plus_dXi = H( X_plus_dXi )
        2/ On calcule la valeur centrale HX = H(X)
        3/ On effectue les différences (HX_plus_dXi-HX) et on divise par
           le pas dXi
        4/ Chaque résultat, par composante, devient une colonne de la Jacobienne

        """
        logging.debug("FDA Début du calcul de la Jacobienne")
        logging.debug("FDA   Incrément de............: %s*X"%float(self.__increment))
        logging.debug("FDA   Approximation centrée...: %s"%(self.__centeredDF))
        #
        if X is None or len(X)==0:
            raise ValueError("Nominal point X for approximate derivatives can not be None or void (given X: %s)."%(str(X),))
        #
        _X = numpy.ravel( X )
        #
        if self.__dX is None:
            _dX  = self.__increment * _X
        else:
            _dX = numpy.ravel( self.__dX )
        assert len(_X) == len(_dX), "Inconsistent dX increment length with respect to the X one"
        assert _X.size == _dX.size, "Inconsistent dX increment size with respect to the X one"
        #
        if (_dX == 0.).any():
            moyenne = _dX.mean()
            if moyenne == 0.:
                _dX = numpy.where( _dX == 0., float(self.__increment), _dX )
            else:
                _dX = numpy.where( _dX == 0., moyenne, _dX )
        #
        __alreadyCalculated  = False
        if self.__avoidRC:
            __bidon, __alreadyCalculatedP = self.__doublon__(_X,  self.__listJPCP, self.__listJPPN, None)
            __bidon, __alreadyCalculatedI = self.__doublon__(_dX, self.__listJPCI, self.__listJPIN, None)
            if __alreadyCalculatedP == __alreadyCalculatedI > -1:
                __alreadyCalculated, __i = True, __alreadyCalculatedP
                logging.debug("FDA Cas J déjà calculé, récupération du doublon %i"%__i)
        #
        if __alreadyCalculated:
            logging.debug("FDA   Calcul Jacobienne (par récupération du doublon %i)"%__i)
            _Jacobienne = self.__listJPCR[__i]
            logging.debug("FDA Fin du calcul de la Jacobienne")
            if dotWith is not None:
                return numpy.dot(_Jacobienne,   numpy.ravel( dotWith ))
            elif dotTWith is not None:
                return numpy.dot(_Jacobienne.T, numpy.ravel( dotTWith ))
        else:
            logging.debug("FDA   Calcul Jacobienne (explicite)")
            if self.__centeredDF:
                #
                if self.__mpEnabled and not self.__mfEnabled:
                    funcrepr = {
                        "__userFunction__path" : self.__userFunction__path,
                        "__userFunction__modl" : self.__userFunction__modl,
                        "__userFunction__name" : self.__userFunction__name,
                    }
                    _jobs = []
                    for i in range( len(_dX) ):
                        _dXi            = _dX[i]
                        _X_plus_dXi     = numpy.array( _X, dtype=float )
                        _X_plus_dXi[i]  = _X[i] + _dXi
                        _X_moins_dXi    = numpy.array( _X, dtype=float )
                        _X_moins_dXi[i] = _X[i] - _dXi
                        #
                        _jobs.append( (_X_plus_dXi,  self.__extraArgs, funcrepr) )
                        _jobs.append( (_X_moins_dXi, self.__extraArgs, funcrepr) )
                    #
                    import multiprocessing
                    self.__pool = multiprocessing.Pool(self.__mpWorkers)
                    _HX_plusmoins_dX = self.__pool.map( ExecuteFunction, _jobs )
                    self.__pool.close()
                    self.__pool.join()
                    #
                    _Jacobienne  = []
                    for i in range( len(_dX) ):
                        _Jacobienne.append( numpy.ravel( _HX_plusmoins_dX[2*i] - _HX_plusmoins_dX[2*i+1] ) / (2.*_dX[i]) )
                    #
                elif self.__mfEnabled:
                    _xserie = []
                    for i in range( len(_dX) ):
                        _dXi            = _dX[i]
                        _X_plus_dXi     = numpy.array( _X, dtype=float )
                        _X_plus_dXi[i]  = _X[i] + _dXi
                        _X_moins_dXi    = numpy.array( _X, dtype=float )
                        _X_moins_dXi[i] = _X[i] - _dXi
                        #
                        _xserie.append( _X_plus_dXi )
                        _xserie.append( _X_moins_dXi )
                    #
                    _HX_plusmoins_dX = self.DirectOperator( _xserie )
                    #
                    _Jacobienne  = []
                    for i in range( len(_dX) ):
                        _Jacobienne.append( numpy.ravel( _HX_plusmoins_dX[2*i] - _HX_plusmoins_dX[2*i+1] ) / (2.*_dX[i]) )
                    #
                else:
                    _Jacobienne  = []
                    for i in range( _dX.size ):
                        _dXi            = _dX[i]
                        _X_plus_dXi     = numpy.array( _X, dtype=float )
                        _X_plus_dXi[i]  = _X[i] + _dXi
                        _X_moins_dXi    = numpy.array( _X, dtype=float )
                        _X_moins_dXi[i] = _X[i] - _dXi
                        #
                        _HX_plus_dXi    = self.DirectOperator( _X_plus_dXi )
                        _HX_moins_dXi   = self.DirectOperator( _X_moins_dXi )
                        #
                        _Jacobienne.append( numpy.ravel( _HX_plus_dXi - _HX_moins_dXi ) / (2.*_dXi) )
                #
            else:
                #
                if self.__mpEnabled and not self.__mfEnabled:
                    funcrepr = {
                        "__userFunction__path" : self.__userFunction__path,
                        "__userFunction__modl" : self.__userFunction__modl,
                        "__userFunction__name" : self.__userFunction__name,
                    }
                    _jobs = []
                    _jobs.append( (_X, self.__extraArgs, funcrepr) )
                    for i in range( len(_dX) ):
                        _X_plus_dXi    = numpy.array( _X, dtype=float )
                        _X_plus_dXi[i] = _X[i] + _dX[i]
                        #
                        _jobs.append( (_X_plus_dXi, self.__extraArgs, funcrepr) )
                    #
                    import multiprocessing
                    self.__pool = multiprocessing.Pool(self.__mpWorkers)
                    _HX_plus_dX = self.__pool.map( ExecuteFunction, _jobs )
                    self.__pool.close()
                    self.__pool.join()
                    #
                    _HX = _HX_plus_dX.pop(0)
                    #
                    _Jacobienne = []
                    for i in range( len(_dX) ):
                        _Jacobienne.append( numpy.ravel(( _HX_plus_dX[i] - _HX ) / _dX[i]) )
                    #
                elif self.__mfEnabled:
                    _xserie = []
                    _xserie.append( _X )
                    for i in range( len(_dX) ):
                        _X_plus_dXi    = numpy.array( _X, dtype=float )
                        _X_plus_dXi[i] = _X[i] + _dX[i]
                        #
                        _xserie.append( _X_plus_dXi )
                    #
                    _HX_plus_dX = self.DirectOperator( _xserie )
                    #
                    _HX = _HX_plus_dX.pop(0)
                    #
                    _Jacobienne = []
                    for i in range( len(_dX) ):
                        _Jacobienne.append( numpy.ravel(( _HX_plus_dX[i] - _HX ) / _dX[i]) )
                    #
                else:
                    _Jacobienne  = []
                    _HX = self.DirectOperator( _X )
                    for i in range( _dX.size ):
                        _dXi            = _dX[i]
                        _X_plus_dXi     = numpy.array( _X, dtype=float )
                        _X_plus_dXi[i]  = _X[i] + _dXi
                        #
                        _HX_plus_dXi = self.DirectOperator( _X_plus_dXi )
                        #
                        _Jacobienne.append( numpy.ravel(( _HX_plus_dXi - _HX ) / _dXi) )
            #
            if (dotWith is not None) or (dotTWith is not None):
                __Produit = self.__listdotwith__(_Jacobienne, dotWith, dotTWith)
            else:
                __Produit = None
            if __Produit is None or self.__avoidRC:
                _Jacobienne = numpy.transpose( numpy.vstack( _Jacobienne ) )
                if self.__avoidRC:
                    if self.__lenghtRJ < 0: self.__lenghtRJ = 2 * _X.size
                    while len(self.__listJPCP) > self.__lenghtRJ:
                        self.__listJPCP.pop(0)
                        self.__listJPCI.pop(0)
                        self.__listJPCR.pop(0)
                        self.__listJPPN.pop(0)
                        self.__listJPIN.pop(0)
                    self.__listJPCP.append( copy.copy(_X) )
                    self.__listJPCI.append( copy.copy(_dX) )
                    self.__listJPCR.append( copy.copy(_Jacobienne) )
                    self.__listJPPN.append( numpy.linalg.norm(_X) )
                    self.__listJPIN.append( numpy.linalg.norm(_Jacobienne) )
            logging.debug("FDA Fin du calcul de la Jacobienne")
            if __Produit is not None:
                return __Produit
        #
        return _Jacobienne

    # ---------------------------------------------------------
    def TangentOperator(self, paire, **extraArgs ):
        """
        Calcul du tangent à l'aide de la Jacobienne.

        NB : les extraArgs sont là pour assurer la compatibilité d'appel, mais
        ne doivent pas être données ici à la fonction utilisateur.
        """
        if self.__mfEnabled:
            assert len(paire) == 1, "Incorrect length of arguments"
            _paire = paire[0]
            assert len(_paire) == 2, "Incorrect number of arguments"
        else:
            assert len(paire) == 2, "Incorrect number of arguments"
            _paire = paire
        X, dX = _paire
        if dX is None or len(dX) == 0:
            #
            # Calcul de la forme matricielle si le second argument est None
            # -------------------------------------------------------------
            _Jacobienne = self.TangentMatrix( X )
            if self.__mfEnabled: return [_Jacobienne,]
            else:                return _Jacobienne
        else:
            #
            # Calcul de la valeur linéarisée de H en X appliqué à dX
            # ------------------------------------------------------
            _HtX = self.TangentMatrix( X, dotWith = dX )
            if self.__mfEnabled: return [_HtX,]
            else:                return _HtX

    # ---------------------------------------------------------
    def AdjointOperator(self, paire, **extraArgs ):
        """
        Calcul de l'adjoint à l'aide de la Jacobienne.

        NB : les extraArgs sont là pour assurer la compatibilité d'appel, mais
        ne doivent pas être données ici à la fonction utilisateur.
        """
        if self.__mfEnabled:
            assert len(paire) == 1, "Incorrect length of arguments"
            _paire = paire[0]
            assert len(_paire) == 2, "Incorrect number of arguments"
        else:
            assert len(paire) == 2, "Incorrect number of arguments"
            _paire = paire
        X, Y = _paire
        if Y is None or len(Y) == 0:
            #
            # Calcul de la forme matricielle si le second argument est None
            # -------------------------------------------------------------
            _JacobienneT = self.TangentMatrix( X ).T
            if self.__mfEnabled: return [_JacobienneT,]
            else:                return _JacobienneT
        else:
            #
            # Calcul de la valeur de l'adjoint en X appliqué à Y
            # --------------------------------------------------
            _HaY = self.TangentMatrix( X, dotTWith = Y )
            if self.__mfEnabled: return [_HaY,]
            else:                return _HaY

# ==============================================================================
def SetInitialDirection( __Direction = [], __Amplitude = 1., __Position = None ):
    "Établit ou élabore une direction avec une amplitude"
    #
    if len(__Direction) == 0 and __Position is None:
        raise ValueError("If initial direction is void, current position has to be given")
    if abs(float(__Amplitude)) < mpr:
        raise ValueError("Amplitude of perturbation can not be zero")
    #
    if len(__Direction) > 0:
        __dX0 = numpy.asarray(__Direction)
    else:
        __dX0 = []
        __X0 = numpy.ravel(numpy.asarray(__Position))
        __mX0 = numpy.mean( __X0, dtype=mfp )
        if abs(__mX0) < 2*mpr: __mX0 = 1. # Évite le problème de position nulle
        for v in __X0:
            if abs(v) > 1.e-8:
                __dX0.append( numpy.random.normal(0.,abs(v)) )
            else:
                __dX0.append( numpy.random.normal(0.,__mX0) )
    #
    __dX0 = numpy.asarray(__dX0,float) # Évite le problème d'array de taille 1
    __dX0 = numpy.ravel( __dX0 )       # Redresse les vecteurs
    __dX0 = float(__Amplitude) * __dX0
    #
    return __dX0

# ==============================================================================
def EnsembleOfCenteredPerturbations( __bgCenter, __bgCovariance, __nbMembers ):
    "Génération d'un ensemble de taille __nbMembers-1 d'états aléatoires centrés"
    #
    __bgCenter = numpy.ravel(__bgCenter)[:,None]
    if __nbMembers < 1:
        raise ValueError("Number of members has to be strictly more than 1 (given number: %s)."%(str(__nbMembers),))
    #
    if __bgCovariance is None:
        _Perturbations = numpy.tile( __bgCenter, __nbMembers)
    else:
        _Z = numpy.random.multivariate_normal(numpy.zeros(__bgCenter.size), __bgCovariance, size=__nbMembers).T
        _Perturbations = numpy.tile( __bgCenter, __nbMembers) + _Z
    #
    return _Perturbations

# ==============================================================================
def EnsembleOfBackgroundPerturbations( __bgCenter, __bgCovariance, __nbMembers, __withSVD = True):
    "Génération d'un ensemble de taille __nbMembers-1 d'états aléatoires centrés"
    def __CenteredRandomAnomalies(Zr, N):
        """
        Génère une matrice de N anomalies aléatoires centrées sur Zr selon les
        notes manuscrites de MB et conforme au code de PS avec eps = -1
        """
        eps = -1
        Q = numpy.identity(N-1)-numpy.ones((N-1,N-1))/numpy.sqrt(N)/(numpy.sqrt(N)-eps)
        Q = numpy.concatenate((Q, [eps*numpy.ones(N-1)/numpy.sqrt(N)]), axis=0)
        R, _ = numpy.linalg.qr(numpy.random.normal(size = (N-1,N-1)))
        Q = numpy.dot(Q,R)
        Zr = numpy.dot(Q,Zr)
        return Zr.T
    #
    __bgCenter = numpy.ravel(__bgCenter).reshape((-1,1))
    if __nbMembers < 1:
        raise ValueError("Number of members has to be strictly more than 1 (given number: %s)."%(str(__nbMembers),))
    if __bgCovariance is None:
        _Perturbations = numpy.tile( __bgCenter, __nbMembers)
    else:
        if __withSVD:
            _U, _s, _V = numpy.linalg.svd(__bgCovariance, full_matrices=False)
            _nbctl = __bgCenter.size
            if __nbMembers > _nbctl:
                _Z = numpy.concatenate((numpy.dot(
                    numpy.diag(numpy.sqrt(_s[:_nbctl])), _V[:_nbctl]),
                    numpy.random.multivariate_normal(numpy.zeros(_nbctl),__bgCovariance,__nbMembers-1-_nbctl)), axis = 0)
            else:
                _Z = numpy.dot(numpy.diag(numpy.sqrt(_s[:__nbMembers-1])), _V[:__nbMembers-1])
            _Zca = __CenteredRandomAnomalies(_Z, __nbMembers)
            _Perturbations = __bgCenter + _Zca
        else:
            if max(abs(__bgCovariance.flatten())) > 0:
                _nbctl = __bgCenter.size
                _Z = numpy.random.multivariate_normal(numpy.zeros(_nbctl),__bgCovariance,__nbMembers-1)
                _Zca = __CenteredRandomAnomalies(_Z, __nbMembers)
                _Perturbations = __bgCenter + _Zca
            else:
                _Perturbations = numpy.tile( __bgCenter, __nbMembers)
    #
    return _Perturbations

# ==============================================================================
def EnsembleMean( __Ensemble ):
    "Renvoie la moyenne empirique d'un ensemble"
    return numpy.asarray(__Ensemble).mean(axis=1, dtype=mfp).astype('float').reshape((-1,1))

# ==============================================================================
def EnsembleOfAnomalies( __Ensemble, __OptMean = None, __Normalisation = 1.):
    "Renvoie les anomalies centrées à partir d'un ensemble"
    if __OptMean is None:
        __Em = EnsembleMean( __Ensemble )
    else:
        __Em = numpy.ravel( __OptMean ).reshape((-1,1))
    #
    return __Normalisation * (numpy.asarray( __Ensemble ) - __Em)

# ==============================================================================
def EnsembleErrorCovariance( __Ensemble, __Quick = False ):
    "Renvoie l'estimation empirique de la covariance d'ensemble"
    if __Quick:
        # Covariance rapide mais rarement définie positive
        __Covariance = numpy.cov( __Ensemble )
    else:
        # Résultat souvent identique à numpy.cov, mais plus robuste
        __n, __m = numpy.asarray( __Ensemble ).shape
        __Anomalies = EnsembleOfAnomalies( __Ensemble )
        # Estimation empirique
        __Covariance = ( __Anomalies @ __Anomalies.T ) / (__m-1)
        # Assure la symétrie
        __Covariance = ( __Covariance + __Covariance.T ) * 0.5
        # Assure la positivité
        __epsilon    = mpr*numpy.trace( __Covariance )
        __Covariance = __Covariance + __epsilon * numpy.identity(__n)
    #
    return __Covariance

# ==============================================================================
def EnsemblePerturbationWithGivenCovariance(
        __Ensemble,
        __Covariance,
        __Seed = None,
        ):
    "Ajout d'une perturbation à chaque membre d'un ensemble selon une covariance prescrite"
    if hasattr(__Covariance,"assparsematrix"):
        if (abs(__Ensemble).mean() > mpr) and (abs(__Covariance.assparsematrix())/abs(__Ensemble).mean() < mpr).all():
            # Traitement d'une covariance nulle ou presque
            return __Ensemble
        if (abs(__Ensemble).mean() <= mpr) and (abs(__Covariance.assparsematrix()) < mpr).all():
            # Traitement d'une covariance nulle ou presque
            return __Ensemble
    else:
        if (abs(__Ensemble).mean() > mpr) and (abs(__Covariance)/abs(__Ensemble).mean() < mpr).all():
            # Traitement d'une covariance nulle ou presque
            return __Ensemble
        if (abs(__Ensemble).mean() <= mpr) and (abs(__Covariance) < mpr).all():
            # Traitement d'une covariance nulle ou presque
            return __Ensemble
    #
    __n, __m = __Ensemble.shape
    if __Seed is not None: numpy.random.seed(__Seed)
    #
    if hasattr(__Covariance,"isscalar") and __Covariance.isscalar():
        # Traitement d'une covariance multiple de l'identité
        __zero = 0.
        __std  = numpy.sqrt(__Covariance.assparsematrix())
        __Ensemble += numpy.random.normal(__zero, __std, size=(__m,__n)).T
    #
    elif hasattr(__Covariance,"isvector") and __Covariance.isvector():
        # Traitement d'une covariance diagonale avec variances non identiques
        __zero = numpy.zeros(__n)
        __std  = numpy.sqrt(__Covariance.assparsematrix())
        __Ensemble += numpy.asarray([numpy.random.normal(__zero, __std) for i in range(__m)]).T
    #
    elif hasattr(__Covariance,"ismatrix") and __Covariance.ismatrix():
        # Traitement d'une covariance pleine
        __Ensemble += numpy.random.multivariate_normal(numpy.zeros(__n), __Covariance.asfullmatrix(__n), size=__m).T
    #
    elif isinstance(__Covariance, numpy.ndarray):
        # Traitement d'une covariance numpy pleine, sachant qu'on arrive ici en dernier
        __Ensemble += numpy.random.multivariate_normal(numpy.zeros(__n), __Covariance, size=__m).T
    #
    else:
        raise ValueError("Error in ensemble perturbation with inadequate covariance specification")
    #
    return __Ensemble

# ==============================================================================
def CovarianceInflation(
        __InputCovOrEns,
        __InflationType   = None,
        __InflationFactor = None,
        __BackgroundCov   = None,
        ):
    """
    Inflation applicable soit sur Pb ou Pa, soit sur les ensembles EXb ou EXa

    Synthèse : Hunt 2007, section 2.3.5
    """
    if __InflationFactor is None:
        return __InputCovOrEns
    else:
        __InflationFactor = float(__InflationFactor)
    #
    __InputCovOrEns = numpy.asarray(__InputCovOrEns)
    if __InputCovOrEns.size == 0: return __InputCovOrEns
    #
    if __InflationType in ["MultiplicativeOnAnalysisCovariance", "MultiplicativeOnBackgroundCovariance"]:
        if __InflationFactor < 1.:
            raise ValueError("Inflation factor for multiplicative inflation has to be greater or equal than 1.")
        if __InflationFactor < 1.+mpr: # No inflation = 1
            return __InputCovOrEns
        __OutputCovOrEns = __InflationFactor**2 * __InputCovOrEns
    #
    elif __InflationType in ["MultiplicativeOnAnalysisAnomalies", "MultiplicativeOnBackgroundAnomalies"]:
        if __InflationFactor < 1.:
            raise ValueError("Inflation factor for multiplicative inflation has to be greater or equal than 1.")
        if __InflationFactor < 1.+mpr: # No inflation = 1
            return __InputCovOrEns
        __InputCovOrEnsMean = __InputCovOrEns.mean(axis=1, dtype=mfp).astype('float')
        __OutputCovOrEns = __InputCovOrEnsMean[:,numpy.newaxis] \
            + __InflationFactor * (__InputCovOrEns - __InputCovOrEnsMean[:,numpy.newaxis])
    #
    elif __InflationType in ["AdditiveOnAnalysisCovariance", "AdditiveOnBackgroundCovariance"]:
        if __InflationFactor < 0.:
            raise ValueError("Inflation factor for additive inflation has to be greater or equal than 0.")
        if __InflationFactor < mpr: # No inflation = 0
            return __InputCovOrEns
        __n, __m = __InputCovOrEns.shape
        if __n != __m:
            raise ValueError("Additive inflation can only be applied to squared (covariance) matrix.")
        __tr = __InputCovOrEns.trace()/__n
        if __InflationFactor > __tr:
            raise ValueError("Inflation factor for additive inflation has to be small over %.0e."%__tr)
        __OutputCovOrEns = (1. - __InflationFactor)*__InputCovOrEns + __InflationFactor * numpy.identity(__n)
    #
    elif __InflationType == "HybridOnBackgroundCovariance":
        if __InflationFactor < 0.:
            raise ValueError("Inflation factor for hybrid inflation has to be greater or equal than 0.")
        if __InflationFactor < mpr: # No inflation = 0
            return __InputCovOrEns
        __n, __m = __InputCovOrEns.shape
        if __n != __m:
            raise ValueError("Additive inflation can only be applied to squared (covariance) matrix.")
        if __BackgroundCov is None:
            raise ValueError("Background covariance matrix B has to be given for hybrid inflation.")
        if __InputCovOrEns.shape != __BackgroundCov.shape:
            raise ValueError("Ensemble covariance matrix has to be of same size than background covariance matrix B.")
        __OutputCovOrEns = (1. - __InflationFactor) * __InputCovOrEns + __InflationFactor * __BackgroundCov
    #
    elif __InflationType == "Relaxation":
        raise NotImplementedError("Relaxation inflation type not implemented")
    #
    else:
        raise ValueError("Error in inflation type, '%s' is not a valid keyword."%__InflationType)
    #
    return __OutputCovOrEns

# ==============================================================================
def HessienneEstimation(__selfA, __nb, __HaM, __HtM, __BI, __RI):
    "Estimation de la Hessienne"
    #
    __HessienneI = []
    for i in range(int(__nb)):
        __ee    = numpy.zeros((__nb,1))
        __ee[i] = 1.
        __HtEE  = numpy.dot(__HtM,__ee).reshape((-1,1))
        __HessienneI.append( numpy.ravel( __BI * __ee + __HaM * (__RI * __HtEE) ) )
    #
    __A = numpy.linalg.inv(numpy.array( __HessienneI ))
    __A = (__A + __A.T) * 0.5 # Symétrie
    __A = __A + mpr*numpy.trace( __A ) * numpy.identity(__nb) # Positivité
    #
    if min(__A.shape) != max(__A.shape):
        raise ValueError(
            "The %s a posteriori covariance matrix A"%(__selfA._name,)+\
            " is of shape %s, despites it has to be a"%(str(__A.shape),)+\
            " squared matrix. There is an error in the observation operator,"+\
            " please check it.")
    if (numpy.diag(__A) < 0).any():
        raise ValueError(
            "The %s a posteriori covariance matrix A"%(__selfA._name,)+\
            " has at least one negative value on its diagonal. There is an"+\
            " error in the observation operator, please check it.")
    if logging.getLogger().level < logging.WARNING: # La vérification n'a lieu qu'en debug
        try:
            numpy.linalg.cholesky( __A )
        except Exception:
            raise ValueError(
                "The %s a posteriori covariance matrix A"%(__selfA._name,)+\
                " is not symmetric positive-definite. Please check your a"+\
                " priori covariances and your observation operator.")
    #
    return __A

# ==============================================================================
def QuantilesEstimations(selfA, A, Xa, HXa = None, Hm = None, HtM = None):
    "Estimation des quantiles a posteriori à partir de A>0 (selfA est modifié)"
    nbsamples = selfA._parameters["NumberOfSamplesForQuantiles"]
    #
    # Traitement des bornes
    if "StateBoundsForQuantiles" in selfA._parameters:
        LBounds = selfA._parameters["StateBoundsForQuantiles"] # Prioritaire
    elif "Bounds" in selfA._parameters:
        LBounds = selfA._parameters["Bounds"]  # Défaut raisonnable
    else:
        LBounds = None
    if LBounds is not None:
        LBounds = ForceNumericBounds( LBounds )
    __Xa = numpy.ravel(Xa)
    #
    # Échantillonnage des états
    YfQ  = None
    EXr  = None
    for i in range(nbsamples):
        if selfA._parameters["SimulationForQuantiles"] == "Linear" and HtM is not None and HXa is not None:
            dXr = (numpy.random.multivariate_normal(__Xa,A) - __Xa).reshape((-1,1))
            if LBounds is not None: # "EstimateProjection" par défaut
                dXr = numpy.max(numpy.hstack((dXr,LBounds[:,0].reshape((-1,1))) - __Xa.reshape((-1,1))),axis=1)
                dXr = numpy.min(numpy.hstack((dXr,LBounds[:,1].reshape((-1,1))) - __Xa.reshape((-1,1))),axis=1)
            dYr = HtM @ dXr
            Yr = HXa.reshape((-1,1)) + dYr
            if selfA._toStore("SampledStateForQuantiles"): Xr = __Xa + numpy.ravel(dXr)
        elif selfA._parameters["SimulationForQuantiles"] == "NonLinear" and Hm is not None:
            Xr = numpy.random.multivariate_normal(__Xa,A)
            if LBounds is not None: # "EstimateProjection" par défaut
                Xr = numpy.max(numpy.hstack((Xr.reshape((-1,1)),LBounds[:,0].reshape((-1,1)))),axis=1)
                Xr = numpy.min(numpy.hstack((Xr.reshape((-1,1)),LBounds[:,1].reshape((-1,1)))),axis=1)
            Yr = numpy.asarray(Hm( Xr ))
        else:
            raise ValueError("Quantile simulations has only to be Linear or NonLinear.")
        #
        if YfQ is None:
            YfQ = Yr.reshape((-1,1))
            if selfA._toStore("SampledStateForQuantiles"): EXr = Xr.reshape((-1,1))
        else:
            YfQ = numpy.hstack((YfQ,Yr.reshape((-1,1))))
            if selfA._toStore("SampledStateForQuantiles"): EXr = numpy.hstack((EXr,Xr.reshape((-1,1))))
    #
    # Extraction des quantiles
    YfQ.sort(axis=-1)
    YQ = None
    for quantile in selfA._parameters["Quantiles"]:
        if not (0. <= float(quantile) <= 1.): continue
        indice = int(nbsamples * float(quantile) - 1./nbsamples)
        if YQ is None: YQ = YfQ[:,indice].reshape((-1,1))
        else:          YQ = numpy.hstack((YQ,YfQ[:,indice].reshape((-1,1))))
    if YQ is not None: # Liste non vide de quantiles
        selfA.StoredVariables["SimulationQuantiles"].store( YQ )
    if selfA._toStore("SampledStateForQuantiles"):
        selfA.StoredVariables["SampledStateForQuantiles"].store( EXr )
    #
    return 0

# ==============================================================================
def ForceNumericBounds( __Bounds, __infNumbers = True ):
    "Force les bornes à être des valeurs numériques, sauf si globalement None"
    # Conserve une valeur par défaut à None s'il n'y a pas de bornes
    if __Bounds is None: return None
    # Converti toutes les bornes individuelles None à +/- l'infini chiffré
    __Bounds = numpy.asarray( __Bounds, dtype=float )
    if len(__Bounds.shape) != 2 or min(__Bounds.shape) <= 0 or __Bounds.shape[1] != 2:
        raise ValueError("Incorrectly shaped bounds data")
    if __infNumbers:
        __Bounds[numpy.isnan(__Bounds[:,0]),0] = -float('inf')
        __Bounds[numpy.isnan(__Bounds[:,1]),1] =  float('inf')
    else:
        __Bounds[numpy.isnan(__Bounds[:,0]),0] = -sys.float_info.max
        __Bounds[numpy.isnan(__Bounds[:,1]),1] =  sys.float_info.max
    return __Bounds

# ==============================================================================
def RecentredBounds( __Bounds, __Center, __Scale = None):
    "Recentre les bornes autour de 0, sauf si globalement None"
    # Conserve une valeur par défaut à None s'il n'y a pas de bornes
    if __Bounds is None: return None
    if __Scale is None:
        # Recentre les valeurs numériques de bornes
        return ForceNumericBounds( __Bounds ) - numpy.ravel( __Center ).reshape((-1,1))
    else:
        # Recentre les valeurs numériques de bornes et change l'échelle par une matrice
        return __Scale @ (ForceNumericBounds( __Bounds, False ) - numpy.ravel( __Center ).reshape((-1,1)))

# ==============================================================================
def ApplyBounds( __Vector, __Bounds, __newClip = True):
    "Applique des bornes numériques à un point"
    # Conserve une valeur par défaut s'il n'y a pas de bornes
    if __Bounds is None: return __Vector
    #
    if not isinstance(__Vector, numpy.ndarray): # Is an array
        raise ValueError("Incorrect array definition of vector data")
    if not isinstance(__Bounds, numpy.ndarray): # Is an array
        raise ValueError("Incorrect array definition of bounds data")
    if 2*__Vector.size != __Bounds.size: # Is a 2 column array of vector lenght
        raise ValueError("Incorrect bounds number (%i) to be applied for this vector (of size %i)"%(__Bounds.size,__Vector.size))
    if len(__Bounds.shape) != 2 or min(__Bounds.shape) <= 0 or __Bounds.shape[1] != 2:
        raise ValueError("Incorrectly shaped bounds data")
    #
    if __newClip:
        __Vector = __Vector.clip(
            __Bounds[:,0].reshape(__Vector.shape),
            __Bounds[:,1].reshape(__Vector.shape),
            )
    else:
        __Vector = numpy.max(numpy.hstack((__Vector.reshape((-1,1)),numpy.asmatrix(__Bounds)[:,0])),axis=1)
        __Vector = numpy.min(numpy.hstack((__Vector.reshape((-1,1)),numpy.asmatrix(__Bounds)[:,1])),axis=1)
        __Vector = numpy.asarray(__Vector)
    #
    return __Vector

# ==============================================================================
def Apply3DVarRecentringOnEnsemble(__EnXn, __EnXf, __Ynpu, __HO, __R, __B, __SuppPars):
    "Recentre l'ensemble Xn autour de l'analyse 3DVAR"
    __Betaf = __SuppPars["HybridCovarianceEquilibrium"]
    #
    Xf = EnsembleMean( __EnXf )
    Pf = Covariance( asCovariance=EnsembleErrorCovariance(__EnXf) )
    Pf = (1 - __Betaf) * __B.asfullmatrix(Xf.size) + __Betaf * Pf
    #
    selfB = PartialAlgorithm("3DVAR")
    selfB._parameters["Minimizer"] = "LBFGSB"
    selfB._parameters["MaximumNumberOfIterations"] = __SuppPars["HybridMaximumNumberOfIterations"]
    selfB._parameters["CostDecrementTolerance"] = __SuppPars["HybridCostDecrementTolerance"]
    selfB._parameters["ProjectedGradientTolerance"] = -1
    selfB._parameters["GradientNormTolerance"] = 1.e-05
    selfB._parameters["StoreInternalVariables"] = False
    selfB._parameters["optiprint"] = -1
    selfB._parameters["optdisp"] = 0
    selfB._parameters["Bounds"] = None
    selfB._parameters["InitializationPoint"] = Xf
    from daAlgorithms.Atoms import std3dvar
    std3dvar.std3dvar(selfB, Xf, __Ynpu, None, __HO, None, __R, Pf)
    Xa = selfB.get("Analysis")[-1].reshape((-1,1))
    del selfB
    #
    return Xa + EnsembleOfAnomalies( __EnXn )

# ==============================================================================
def multiXOsteps(selfA, Xb, Y, U, HO, EM, CM, R, B, Q, oneCycle,
        __CovForecast = False, __LinEvolution = False,
        ):
    """
    Prévision multi-pas avec une correction par pas (multi-méthodes)
    """
    #
    # Initialisation
    # --------------
    if selfA._parameters["EstimationOf"] == "State":
        if len(selfA.StoredVariables["Analysis"])==0 or not selfA._parameters["nextStep"]:
            Xn = numpy.asarray(Xb)
            if __CovForecast: Pn = B
            selfA.StoredVariables["Analysis"].store( Xn )
            if selfA._toStore("APosterioriCovariance"):
                if hasattr(B,"asfullmatrix"):
                    selfA.StoredVariables["APosterioriCovariance"].store( B.asfullmatrix(Xn.size) )
                else:
                    selfA.StoredVariables["APosterioriCovariance"].store( B )
            selfA._setInternalState("seed", numpy.random.get_state())
        elif selfA._parameters["nextStep"]:
            Xn = selfA._getInternalState("Xn")
            if __CovForecast: Pn = selfA._getInternalState("Pn")
    else:
        Xn = numpy.asarray(Xb)
        if __CovForecast: Pn = B
    #
    if hasattr(Y,"stepnumber"):
        duration = Y.stepnumber()
    else:
        duration = 2
    #
    # Multi-steps
    # -----------
    for step in range(duration-1):
        selfA.StoredVariables["CurrentStepNumber"].store( len(selfA.StoredVariables["Analysis"]) )
        #
        if hasattr(Y,"store"):
            Ynpu = numpy.asarray( Y[step+1] ).reshape((-1,1))
        else:
            Ynpu = numpy.asarray( Y ).reshape((-1,1))
        #
        if U is not None:
            if hasattr(U,"store") and len(U)>1:
                Un = numpy.asarray( U[step] ).reshape((-1,1))
            elif hasattr(U,"store") and len(U)==1:
                Un = numpy.asarray( U[0] ).reshape((-1,1))
            else:
                Un = numpy.asarray( U ).reshape((-1,1))
        else:
            Un = None
        #
        # Predict (Time Update)
        # ---------------------
        if selfA._parameters["EstimationOf"] == "State":
            if __CovForecast or __LinEvolution:
                Mt = EM["Tangent"].asMatrix(Xn)
                Mt = Mt.reshape(Xn.size,Xn.size) # ADAO & check shape
            if __CovForecast:
                Ma = EM["Adjoint"].asMatrix(Xn)
                Ma = Ma.reshape(Xn.size,Xn.size) # ADAO & check shape
                Pn_predicted = Q + Mt @ (Pn @ Ma)
            if __LinEvolution:
                Xn_predicted = Mt @ Xn
            else:
                M  = EM["Direct"].appliedControledFormTo
                Xn_predicted = M( (Xn, Un) )
            if CM is not None and "Tangent" in CM and Un is not None: # Attention : si Cm est aussi dans M, doublon !
                Cm = CM["Tangent"].asMatrix(Xn_predicted)
                Cm = Cm.reshape(Xn.size,Un.size) # ADAO & check shape
                Xn_predicted = Xn_predicted + (Cm @ Un).reshape((-1,1))
        elif selfA._parameters["EstimationOf"] == "Parameters": # No forecast
            # --- > Par principe, M = Id, Q = 0
            Xn_predicted = Xn
            if __CovForecast: Pn_predicted = Pn
        Xn_predicted = numpy.asarray(Xn_predicted).reshape((-1,1))
        if selfA._toStore("ForecastState"):
            selfA.StoredVariables["ForecastState"].store( Xn_predicted )
        if __CovForecast:
            if hasattr(Pn_predicted,"asfullmatrix"):
                Pn_predicted = Pn_predicted.asfullmatrix(Xn.size)
            else:
                Pn_predicted = numpy.asarray(Pn_predicted).reshape((Xn.size,Xn.size))
            if selfA._toStore("ForecastCovariance"):
                selfA.StoredVariables["ForecastCovariance"].store( Pn_predicted )
        #
        # Correct (Measurement Update)
        # ----------------------------
        if __CovForecast:
            oneCycle(selfA, Xn_predicted, Ynpu, Un, HO, CM, R, Pn_predicted, True)
        else:
            oneCycle(selfA, Xn_predicted, Ynpu, Un, HO, CM, R, B, True)
        #
        #--------------------------
        Xn = selfA._getInternalState("Xn")
        if __CovForecast: Pn = selfA._getInternalState("Pn")
    #
    return 0


# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

"""
    Normalized interface for ADAO scripting (full version API)

    The main interface to use is an object "New" build from "adaoBuilder".

    Usage by an example:

        from numpy import array, matrix
        from adao import adaoBuilder
        case = adaoBuilder.New()
        case.set( 'AlgorithmParameters', Algorithm='3DVAR' )
        case.set( 'Background',          Vector=[0, 1, 2] )
        case.set( 'BackgroundError',     ScalarSparseMatrix=1.0 )
        case.set( 'Observation',         Vector=array([0.5, 1.5, 2.5]) )
        case.set( 'ObservationError',    DiagonalSparseMatrix='1 1 1' )
        case.set( 'ObservationOperator', Matrix='1 0 0;0 2 0;0 0 3' )
        case.execute()
        #
        print(case.get("Analysis")[-1])

    Documentation

        See associated up-to-date documentation for details of commands.
"""
__author__ = "Jean-Philippe ARGAUD"
__all__ = ["New"]

from daCore.Aidsm import Aidsm as _Aidsm
from daCore.version import name, version, year, date, __version__

# ==============================================================================
class New(_Aidsm):
    """
    Generic ADAO TUI builder
    """
    def __init__(self, __name = ""):
        _Aidsm.__init__(self, __name)

class Gui(object):
    """
    Generic ADAO GUI builder
    """
    def __init__(self):
        from daCore.Interfaces import EficasGUI
        EficasGUI().gui()

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC \n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import math, numpy
from daCore import BasicObjects, NumericObjects, PlatformInfo
mpr = PlatformInfo.PlatformInfo().MachinePrecision()

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "GRADIENTTEST")
        self.defineRequiredParameter(
            name     = "ResiduFormula",
            default  = "Taylor",
            typecast = str,
            message  = "Formule de résidu utilisée",
            listval  = ["Norm", "TaylorOnNorm", "Taylor"],
            )
        self.defineRequiredParameter(
            name     = "EpsilonMinimumExponent",
            default  = -8,
            typecast = int,
            message  = "Exposant minimal en puissance de 10 pour le multiplicateur d'incrément",
            minval   = -20,
            maxval   = 0,
            )
        self.defineRequiredParameter(
            name     = "InitialDirection",
            default  = [],
            typecast = list,
            message  = "Direction initiale de la dérivée directionnelle autour du point nominal",
            )
        self.defineRequiredParameter(
            name     = "AmplitudeOfInitialDirection",
            default  = 1.,
            typecast = float,
            message  = "Amplitude de la direction initiale de la dérivée directionnelle autour du point nominal",
            )
        self.defineRequiredParameter(
            name     = "AmplitudeOfTangentPerturbation",
            default  = 1.e-2,
            typecast = float,
            message  = "Amplitude de la perturbation pour le calcul de la forme tangente",
            minval   = 1.e-10,
            maxval   = 1.,
            )
        self.defineRequiredParameter(
            name     = "SetSeed",
            typecast = numpy.random.seed,
            message  = "Graine fixée pour le générateur aléatoire",
            )
        self.defineRequiredParameter(
            name     = "PlotAndSave",
            default  = False,
            typecast = bool,
            message  = "Trace et sauve les résultats",
            )
        self.defineRequiredParameter(
            name     = "ResultFile",
            default  = self._name+"_result_file",
            typecast = str,
            message  = "Nom de base (hors extension) des fichiers de sauvegarde des résultats",
            )
        self.defineRequiredParameter(
            name     = "ResultTitle",
            default  = "",
            typecast = str,
            message  = "Titre du tableau et de la figure",
            )
        self.defineRequiredParameter(
            name     = "ResultLabel",
            default  = "",
            typecast = str,
            message  = "Label de la courbe tracée dans la figure",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "CurrentState",
                "Residu",
                "SimulatedObservationAtCurrentState",
                ]
            )
        self.requireInputArguments(
            mandatory= ("Xb", "HO"),
            )
        self.setAttributes(tags=(
            "Checking",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, U, HO, EM, CM, R, B, Q)
        #
        Hm = HO["Direct"].appliedTo
        if self._parameters["ResiduFormula"] in ["Taylor", "TaylorOnNorm"]:
            Ht = HO["Tangent"].appliedInXTo
        #
        # ----------
        Perturbations = [ 10**i for i in range(self._parameters["EpsilonMinimumExponent"],1) ]
        Perturbations.reverse()
        #
        X       = numpy.ravel(    Xb   ).reshape((-1,1))
        FX      = numpy.ravel( Hm( X ) ).reshape((-1,1))
        NormeX  = numpy.linalg.norm( X )
        NormeFX = numpy.linalg.norm( FX )
        if NormeFX < mpr: NormeFX = mpr
        if self._toStore("CurrentState"):
            self.StoredVariables["CurrentState"].store( X )
        if self._toStore("SimulatedObservationAtCurrentState"):
            self.StoredVariables["SimulatedObservationAtCurrentState"].store( FX )
        #
        dX0 = NumericObjects.SetInitialDirection(
            self._parameters["InitialDirection"],
            self._parameters["AmplitudeOfInitialDirection"],
            X,
            )
        #
        if self._parameters["ResiduFormula"] in ["Taylor", "TaylorOnNorm"]:
            dX1      = float(self._parameters["AmplitudeOfTangentPerturbation"]) * dX0.reshape((-1,1))
            GradFxdX = Ht( (X, dX1) )
            GradFxdX = numpy.ravel( GradFxdX ).reshape((-1,1))
            GradFxdX = float(1./self._parameters["AmplitudeOfTangentPerturbation"]) * GradFxdX
        #
        # Entete des resultats
        # --------------------
        __marge =  12*u" "
        __precision = u"""
            Remarque : les nombres inferieurs a %.0e (environ) representent un zero
                       a la precision machine.\n"""%mpr
        if self._parameters["ResiduFormula"] == "Taylor":
            __entete = u"  i   Alpha       ||X||    ||F(X)||  ||F(X+dX)||    ||dX||  ||F(X+dX)-F(X)||   ||F(X+dX)-F(X)||/||dX||      R(Alpha)   log( R )"
            __msgdoc = u"""
            On observe le residu issu du developpement de Taylor de la fonction F,
            normalise par la valeur au point nominal :

                         || F(X+Alpha*dX) - F(X) - Alpha * GradientF_X(dX) ||
              R(Alpha) = ----------------------------------------------------
                                         || F(X) ||

            Si le residu decroit et que la decroissance se fait en Alpha**2 selon Alpha,
            cela signifie que le gradient est bien calcule jusqu'a la precision d'arret
            de la decroissance quadratique, et que F n'est pas lineaire.

            Si le residu decroit et que la decroissance se fait en Alpha selon Alpha,
            jusqu'a un certain seuil apres lequel le residu est faible et constant, cela
            signifie que F est lineaire et que le residu decroit a partir de l'erreur
            faite dans le calcul du terme GradientF_X.

            On prend dX0 = Normal(0,X) et dX = Alpha*dX0. F est le code de calcul.\n""" + __precision
        if self._parameters["ResiduFormula"] == "TaylorOnNorm":
            __entete = u"  i   Alpha       ||X||    ||F(X)||  ||F(X+dX)||    ||dX||  ||F(X+dX)-F(X)||   ||F(X+dX)-F(X)||/||dX||      R(Alpha)   log( R )"
            __msgdoc = u"""
            On observe le residu issu du developpement de Taylor de la fonction F,
            rapporte au parametre Alpha au carre :

                         || F(X+Alpha*dX) - F(X) - Alpha * GradientF_X(dX) ||
              R(Alpha) = ----------------------------------------------------
                                            Alpha**2

            C'est un residu essentiellement similaire au critere classique de Taylor,
            mais son comportement peut differer selon les proprietes numeriques des
            calculs de ses differents termes.

            Si le residu est constant jusqu'a un certain seuil et croissant ensuite,
            cela signifie que le gradient est bien calcule jusqu'a cette precision
            d'arret, et que F n'est pas lineaire.

            Si le residu est systematiquement croissant en partant d'une valeur faible
            par rapport a ||F(X)||, cela signifie que F est (quasi-)lineaire et que le
            calcul du gradient est correct jusqu'au moment ou le residu est de l'ordre de
            grandeur de ||F(X)||.

            On prend dX0 = Normal(0,X) et dX = Alpha*dX0. F est le code de calcul.\n""" + __precision
        if self._parameters["ResiduFormula"] == "Norm":
            __entete = u"  i   Alpha       ||X||    ||F(X)||  ||F(X+dX)||    ||dX||  ||F(X+dX)-F(X)||   ||F(X+dX)-F(X)||/||dX||      R(Alpha)   log( R )"
            __msgdoc = u"""
            On observe le residu, qui est base sur une approximation du gradient :

                          || F(X+Alpha*dX) - F(X) ||
              R(Alpha) =  ---------------------------
                                    Alpha

            qui doit rester constant jusqu'a ce que l'on atteigne la precision du calcul.

            On prend dX0 = Normal(0,X) et dX = Alpha*dX0. F est le code de calcul.\n""" + __precision
        #
        if len(self._parameters["ResultTitle"]) > 0:
            __rt = str(self._parameters["ResultTitle"])
            msgs  = u"\n"
            msgs += __marge + "====" + "="*len(__rt) + "====\n"
            msgs += __marge + "    " + __rt + "\n"
            msgs += __marge + "====" + "="*len(__rt) + "====\n"
        else:
            msgs  = u""
        msgs += __msgdoc
        #
        __nbtirets = len(__entete) + 2
        msgs += "\n" + __marge + "-"*__nbtirets
        msgs += "\n" + __marge + __entete
        msgs += "\n" + __marge + "-"*__nbtirets
        #
        # Boucle sur les perturbations
        # ----------------------------
        NormesdX     = []
        NormesFXdX   = []
        NormesdFX    = []
        NormesdFXsdX = []
        NormesdFXsAm = []
        NormesdFXGdX = []
        #
        for i,amplitude in enumerate(Perturbations):
            dX      = amplitude * dX0.reshape((-1,1))
            #
            FX_plus_dX = Hm( X + dX )
            FX_plus_dX = numpy.ravel( FX_plus_dX ).reshape((-1,1))
            #
            if self._toStore("CurrentState"):
                self.StoredVariables["CurrentState"].store( numpy.ravel(X + dX) )
            if self._toStore("SimulatedObservationAtCurrentState"):
                self.StoredVariables["SimulatedObservationAtCurrentState"].store( numpy.ravel(FX_plus_dX) )
            #
            NormedX     = numpy.linalg.norm( dX )
            NormeFXdX   = numpy.linalg.norm( FX_plus_dX )
            NormedFX    = numpy.linalg.norm( FX_plus_dX - FX )
            NormedFXsdX = NormedFX/NormedX
            # Residu Taylor
            if self._parameters["ResiduFormula"] in ["Taylor", "TaylorOnNorm"]:
                NormedFXGdX = numpy.linalg.norm( FX_plus_dX - FX - amplitude * GradFxdX )
            # Residu Norm
            NormedFXsAm = NormedFX/amplitude
            #
            # if numpy.abs(NormedFX) < 1.e-20:
            #     break
            #
            NormesdX.append(     NormedX     )
            NormesFXdX.append(   NormeFXdX   )
            NormesdFX.append(    NormedFX    )
            if self._parameters["ResiduFormula"] in ["Taylor", "TaylorOnNorm"]:
                NormesdFXGdX.append( NormedFXGdX )
            NormesdFXsdX.append( NormedFXsdX )
            NormesdFXsAm.append( NormedFXsAm )
            #
            if self._parameters["ResiduFormula"] == "Taylor":
                Residu = NormedFXGdX / NormeFX
            elif self._parameters["ResiduFormula"] == "TaylorOnNorm":
                Residu = NormedFXGdX / (amplitude*amplitude)
            elif self._parameters["ResiduFormula"] == "Norm":
                Residu = NormedFXsAm
            #
            msg = "  %2i  %5.0e   %9.3e   %9.3e   %9.3e   %9.3e   %9.3e      |      %9.3e          |   %9.3e   %4.0f"%(i,amplitude,NormeX,NormeFX,NormeFXdX,NormedX,NormedFX,NormedFXsdX,Residu,math.log10(max(1.e-99,Residu)))
            msgs += "\n" + __marge + msg
            #
            self.StoredVariables["Residu"].store( Residu )
        #
        msgs += "\n" + __marge + "-"*__nbtirets
        msgs += "\n"
        #
        # ----------
        print("\nResults of gradient check by \"%s\" formula:"%self._parameters["ResiduFormula"])
        print(msgs)
        #
        if self._parameters["PlotAndSave"]:
            f = open(str(self._parameters["ResultFile"])+".txt",'a')
            f.write(msgs)
            f.close()
            #
            Residus = self.StoredVariables["Residu"][-len(Perturbations):]
            if self._parameters["ResiduFormula"] in ["Taylor", "TaylorOnNorm"]:
                PerturbationsCarre = [ 10**(2*i) for i in range(-len(NormesdFXGdX)+1,1) ]
                PerturbationsCarre.reverse()
                dessiner(
                    Perturbations,
                    Residus,
                    titre    = self._parameters["ResultTitle"],
                    label    = self._parameters["ResultLabel"],
                    logX     = True,
                    logY     = True,
                    filename = str(self._parameters["ResultFile"])+".ps",
                    YRef     = PerturbationsCarre,
                    normdY0  = numpy.log10( NormesdFX[0] ),
                    )
            elif self._parameters["ResiduFormula"] == "Norm":
                dessiner(
                    Perturbations,
                    Residus,
                    titre    = self._parameters["ResultTitle"],
                    label    = self._parameters["ResultLabel"],
                    logX     = True,
                    logY     = True,
                    filename = str(self._parameters["ResultFile"])+".ps",
                    )
        #
        self._post_run(HO)
        return 0

# ==============================================================================

def dessiner(
        X,
        Y,
        titre     = "",
        label     = "",
        logX      = False,
        logY      = False,
        filename  = "",
        pause     = False,
        YRef      = None, # Vecteur de reference a comparer a Y
        recalYRef = True, # Decalage du point 0 de YRef a Y[0]
        normdY0   = 0.,   # Norme de DeltaY[0]
        ):
    import Gnuplot
    __gnuplot = Gnuplot
    __g = __gnuplot.Gnuplot(persist=1) # persist=1
    # __g('set terminal '+__gnuplot.GnuplotOpts.default_term)
    __g('set style data lines')
    __g('set grid')
    __g('set autoscale')
    __g('set title  "'+titre+'"')
    # __g('set range [] reverse')
    # __g('set yrange [0:2]')
    #
    if logX:
        steps = numpy.log10( X )
        __g('set xlabel "Facteur multiplicatif de dX, en echelle log10"')
    else:
        steps = X
        __g('set xlabel "Facteur multiplicatif de dX"')
    #
    if logY:
        values = numpy.log10( Y )
        __g('set ylabel "Amplitude du residu, en echelle log10"')
    else:
        values = Y
        __g('set ylabel "Amplitude du residu"')
    #
    __g.plot( __gnuplot.Data( steps, values, title=label, with_='lines lw 3' ) )
    if YRef is not None:
        if logY:
            valuesRef = numpy.log10( YRef )
        else:
            valuesRef = YRef
        if recalYRef and not numpy.all(values < -8):
            valuesRef = valuesRef + values[0]
        elif recalYRef and numpy.all(values < -8):
            valuesRef = valuesRef + normdY0
        else:
            pass
        __g.replot( __gnuplot.Data( steps, valuesRef, title="Reference", with_='lines lw 1' ) )
    #
    if filename != "":
        __g.hardcopy( filename, color=1)
    if pause:
        eval(input('Please press return to continue...\n'))

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import numpy
from daCore import BasicObjects

# ==============================================================================
class ElementaryAlgorithm(BasicObjects.Algorithm):
    def __init__(self):
        BasicObjects.Algorithm.__init__(self, "TABUSEARCH")
        self.defineRequiredParameter(
            name     = "MaximumNumberOfIterations",
            default  = 50,
            typecast = int,
            message  = "Nombre maximal de pas d'optimisation",
            minval   = 1,
            oldname  = "MaximumNumberOfSteps",
            )
        self.defineRequiredParameter(
            name     = "SetSeed",
            typecast = numpy.random.seed,
            message  = "Graine fixée pour le générateur aléatoire",
            )
        self.defineRequiredParameter(
            name     = "LengthOfTabuList",
            default  = 50,
            typecast = int,
            message  = "Longueur de la liste tabou",
            minval   = 1,
            )
        self.defineRequiredParameter(
            name     = "NumberOfElementaryPerturbations",
            default  = 1,
            typecast = int,
            message  = "Nombre de perturbations élémentaires pour choisir une perturbation d'état",
            minval   = 1,
            )
        self.defineRequiredParameter(
            name     = "NoiseDistribution",
            default  = "Uniform",
            typecast = str,
            message  = "Distribution pour générer les perturbations d'état",
            listval  = ["Gaussian","Uniform"],
            )
        self.defineRequiredParameter(
            name     = "QualityCriterion",
            default  = "AugmentedWeightedLeastSquares",
            typecast = str,
            message  = "Critère de qualité utilisé",
            listval  = [
                "AugmentedWeightedLeastSquares", "AWLS", "DA",
                "WeightedLeastSquares", "WLS",
                "LeastSquares", "LS", "L2",
                "AbsoluteValue", "L1",
                "MaximumError", "ME",
                ],
            )
        self.defineRequiredParameter(
            name     = "NoiseHalfRange",
            default  = [],
            typecast = numpy.ravel,
            message  = "Demi-amplitude des perturbations uniformes centrées d'état pour chaque composante de l'état",
            )
        self.defineRequiredParameter(
            name     = "StandardDeviation",
            default  = [],
            typecast = numpy.ravel,
            message  = "Ecart-type des perturbations gaussiennes d'état pour chaque composante de l'état",
            )
        self.defineRequiredParameter(
            name     = "NoiseAddingProbability",
            default  = 1.,
            typecast = float,
            message  = "Probabilité de perturbation d'une composante de l'état",
            minval   = 0.,
            maxval   = 1.,
            )
        self.defineRequiredParameter(
            name     = "StoreInternalVariables",
            default  = False,
            typecast = bool,
            message  = "Stockage des variables internes ou intermédiaires du calcul",
            )
        self.defineRequiredParameter(
            name     = "StoreSupplementaryCalculations",
            default  = [],
            typecast = tuple,
            message  = "Liste de calculs supplémentaires à stocker et/ou effectuer",
            listval  = [
                "Analysis",
                "BMA",
                "CostFunctionJ",
                "CostFunctionJb",
                "CostFunctionJo",
                "CurrentIterationNumber",
                "CurrentState",
                "Innovation",
                "OMA",
                "OMB",
                "SimulatedObservationAtBackground",
                "SimulatedObservationAtCurrentState",
                "SimulatedObservationAtOptimum",
                ]
            )
        self.defineRequiredParameter( # Pas de type
            name     = "Bounds",
            message  = "Liste des valeurs de bornes",
            )
        self.requireInputArguments(
            mandatory= ("Xb", "Y", "HO", "R", "B"),
            )
        self.setAttributes(tags=(
            "Optimization",
            "NonLinear",
            "MetaHeuristic",
            ))

    def run(self, Xb=None, Y=None, U=None, HO=None, EM=None, CM=None, R=None, B=None, Q=None, Parameters=None):
        self._pre_run(Parameters, Xb, Y, U, HO, EM, CM, R, B, Q)
        #
        if self._parameters["NoiseDistribution"] == "Uniform":
            nrange = self._parameters["NoiseHalfRange"] # Vecteur
            if nrange.size != Xb.size:
                raise ValueError("Noise generation by Uniform distribution requires range for all variable increments. The actual noise half range vector is:\n%s"%nrange)
        elif self._parameters["NoiseDistribution"] == "Gaussian":
            sigma = numpy.ravel(self._parameters["StandardDeviation"]) # Vecteur
            if sigma.size != Xb.size:
                raise ValueError("Noise generation by Gaussian distribution requires standard deviation for all variable increments. The actual standard deviation vector is:\n%s"%sigma)
        #
        Hm = HO["Direct"].appliedTo
        #
        BI = B.getI()
        RI = R.getI()
        #
        def Tweak( x, NoiseDistribution, NoiseAddingProbability ):
            _X  = numpy.array( x, dtype=float, copy=True ).ravel().reshape((-1,1))
            if NoiseDistribution == "Uniform":
                for i in range(_X.size):
                    if NoiseAddingProbability >= numpy.random.uniform():
                        _increment = numpy.random.uniform(low=-nrange[i], high=nrange[i])
                        # On ne traite pas encore le dépassement des bornes ici
                        _X[i] += _increment
            elif NoiseDistribution == "Gaussian":
                for i in range(_X.size):
                    if NoiseAddingProbability >= numpy.random.uniform():
                        _increment = numpy.random.normal(loc=0., scale=sigma[i])
                        # On ne traite pas encore le dépassement des bornes ici
                        _X[i] += _increment
            #
            return _X
        #
        def StateInList( x, _TL ):
            _X  = numpy.ravel( x )
            _xInList = False
            for state in _TL:
                if numpy.all(numpy.abs( _X - numpy.ravel(state) ) <= 1e-16*numpy.abs(_X)):
                    _xInList = True
            # if _xInList: import sys ; sys.exit()
            return _xInList
        #
        def CostFunction(x, QualityMeasure="AugmentedWeightedLeastSquares"):
            _X  = numpy.ravel( x ).reshape((-1,1))
            _HX = numpy.ravel( Hm( _X ) ).reshape((-1,1))
            _Innovation = Y - _HX
            #
            if QualityMeasure in ["AugmentedWeightedLeastSquares","AWLS","DA"]:
                if BI is None or RI is None:
                    raise ValueError("Background and Observation error covariance matrices has to be properly defined!")
                Jb  = 0.5 * (_X - Xb).T @ (BI @ (_X - Xb))
                Jo  = 0.5 * _Innovation.T @ (RI @ _Innovation)
            elif QualityMeasure in ["WeightedLeastSquares","WLS"]:
                if RI is None:
                    raise ValueError("Observation error covariance matrix has to be properly defined!")
                Jb  = 0.
                Jo  = 0.5 * _Innovation.T @ (RI @ _Innovation)
            elif QualityMeasure in ["LeastSquares","LS","L2"]:
                Jb  = 0.
                Jo  = 0.5 * _Innovation.T @ _Innovation
            elif QualityMeasure in ["AbsoluteValue","L1"]:
                Jb  = 0.
                Jo  = numpy.sum( numpy.abs(_Innovation) )
            elif QualityMeasure in ["MaximumError","ME"]:
                Jb  = 0.
                Jo  = numpy.max( numpy.abs(_Innovation) )
            #
            J   = float( Jb ) + float( Jo )
            #
            return J
        #
        # Minimisation de la fonctionnelle
        # --------------------------------
        _n = 0
        _S = Xb
        _qualityS = CostFunction( _S, self._parameters["QualityCriterion"] )
        _Best, _qualityBest   =   _S, _qualityS
        _TabuList = []
        _TabuList.append( _S )
        while _n < self._parameters["MaximumNumberOfIterations"]:
            _n += 1
            if len(_TabuList) > self._parameters["LengthOfTabuList"]:
                _TabuList.pop(0)
            _R = Tweak( _S, self._parameters["NoiseDistribution"], self._parameters["NoiseAddingProbability"] )
            _qualityR = CostFunction( _R, self._parameters["QualityCriterion"] )
            for nbt in range(self._parameters["NumberOfElementaryPerturbations"]-1):
                _W = Tweak( _S, self._parameters["NoiseDistribution"], self._parameters["NoiseAddingProbability"] )
                _qualityW = CostFunction( _W, self._parameters["QualityCriterion"] )
                if (not StateInList(_W, _TabuList)) and ( (_qualityW < _qualityR) or StateInList(_R,_TabuList) ):
                    _R, _qualityR   =   _W, _qualityW
            if (not StateInList( _R, _TabuList )) and (_qualityR < _qualityS):
                _S, _qualityS   =   _R, _qualityR
                _TabuList.append( _S )
            if _qualityS < _qualityBest:
                _Best, _qualityBest   =   _S, _qualityS
            #
            self.StoredVariables["CurrentIterationNumber"].store( len(self.StoredVariables["CostFunctionJ"]) )
            if self._parameters["StoreInternalVariables"] or self._toStore("CurrentState"):
                self.StoredVariables["CurrentState"].store( _Best )
            if self._toStore("SimulatedObservationAtCurrentState"):
                _HmX = Hm( _Best )
                self.StoredVariables["SimulatedObservationAtCurrentState"].store( _HmX )
            self.StoredVariables["CostFunctionJb"].store( 0. )
            self.StoredVariables["CostFunctionJo"].store( 0. )
            self.StoredVariables["CostFunctionJ" ].store( _qualityBest )
        #
        # Obtention de l'analyse
        # ----------------------
        Xa = _Best
        #
        self.StoredVariables["Analysis"].store( Xa )
        #
        # Calculs et/ou stockages supplémentaires
        # ---------------------------------------
        if self._toStore("OMA") or \
            self._toStore("SimulatedObservationAtOptimum"):
            HXa = Hm(Xa).reshape((-1,1))
        if self._toStore("Innovation") or \
            self._toStore("OMB") or \
            self._toStore("SimulatedObservationAtBackground"):
            HXb = Hm(Xb).reshape((-1,1))
            Innovation = Y - HXb
        if self._toStore("Innovation"):
            self.StoredVariables["Innovation"].store( Innovation )
        if self._toStore("OMB"):
            self.StoredVariables["OMB"].store( Innovation )
        if self._toStore("BMA"):
            self.StoredVariables["BMA"].store( numpy.ravel(Xb) - numpy.ravel(Xa) )
        if self._toStore("OMA"):
            self.StoredVariables["OMA"].store( Y - HXa )
        if self._toStore("SimulatedObservationAtBackground"):
            self.StoredVariables["SimulatedObservationAtBackground"].store( HXb )
        if self._toStore("SimulatedObservationAtOptimum"):
            self.StoredVariables["SimulatedObservationAtOptimum"].store( HXa )
        #
        self._post_run(HO)
        return 0

# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

"""
    Définit les outils d'interfaces normalisées de cas.
"""
__author__ = "Jean-Philippe ARGAUD"
__all__ = []

import os
import sys
import numpy
import mimetypes
import logging
import copy
from daCore import Persistence
from daCore import PlatformInfo
from daCore import Templates
from daCore import Reporting

# ==============================================================================
class GenericCaseViewer(object):
    """
    Gestion des commandes de création d'une vue de cas
    """
    def __init__(self, __name="", __objname="case", __content=None, __object=None):
        "Initialisation et enregistrement de l'entete"
        self._name         = str(__name)
        self._objname      = str(__objname)
        self._lineSerie    = []
        self._switchoff    = False
        self._numobservers = 2
        self._content      = __content
        self._object       = __object
        self._missing = """raise ValueError("This case requires beforehand to import or define the variable named <%s>. When corrected, remove this command, correct and uncomment the following one.")\n# """
    #------------------------
    def _append(self, *args):
        "Transformation d'une commande individuelle en un enregistrement"
        raise NotImplementedError()
    def _extract(self, *args):
        "Transformation d'enregistrement(s) en commande(s) individuelle(s)"
        raise NotImplementedError()
    #------------------------------
    def _initialize(self, __multilines):
        "Permet des pré-conversions automatiques simples de commandes ou clés"
        __translation = {
            "Study_name"          :"StudyName",
            "Study_repertory"     :"StudyRepertory",
            "MaximumNumberOfSteps":"MaximumNumberOfIterations",
            "FunctionDict"        :"ScriptWithSwitch",
            "FUNCTIONDICT_FILE"   :"SCRIPTWITHSWITCH_FILE",
        }
        for k,v in __translation.items():
            __multilines = __multilines.replace(k,v)
        return __multilines
    def _finalize(self, __upa=None):
        "Enregistrement du final"
        __hasNotExecute = True
        for __l in self._lineSerie:
            if "%s.execute"%(self._objname,) in __l: __hasNotExecute = False
        if __hasNotExecute:
            self._lineSerie.append("%s.execute()"%(self._objname,))
        if __upa is not None and len(__upa)>0:
            __upa = __upa.replace("ADD",str(self._objname))
            self._lineSerie.append(__upa)
    def _addLine(self, line=""):
        "Ajoute un enregistrement individuel"
        self._lineSerie.append(line)
    def _get_objname(self):
        return self._objname
    def dump(self, __filename=None, __upa=None):
        "Restitution normalisée des commandes"
        self._finalize(__upa)
        __text = "\n".join(self._lineSerie)
        __text +="\n"
        if __filename is not None:
            __file = os.path.abspath(__filename)
            __fid = open(__file,"w")
            __fid.write(__text)
            __fid.close()
        return __text
    def load(self, __filename=None, __content=None, __object=None):
        "Chargement normalisé des commandes"
        if __filename is not None and os.path.exists(__filename):
            self._content = open(__filename, 'r').read()
            self._content = self._initialize(self._content)
        elif __content is not None and type(__content) is str:
            self._content = self._initialize(__content)
        elif __object is not None and type(__object) is dict:
            self._object = copy.deepcopy(__object)
        else:
            pass # use "self._content" from initialization
        __commands = self._extract(self._content, self._object)
        return __commands

class _TUIViewer(GenericCaseViewer):
    """
    Établissement des commandes d'un cas ADAO TUI (Cas<->TUI)
    """
    def __init__(self, __name="", __objname="case", __content=None, __object=None):
        "Initialisation et enregistrement de l'entete"
        GenericCaseViewer.__init__(self, __name, __objname, __content, __object)
        self._addLine("# -*- coding: utf-8 -*-")
        self._addLine("#\n# Python script using ADAO TUI\n#")
        self._addLine("from numpy import array, matrix")
        self._addLine("from adao import adaoBuilder")
        self._addLine("%s = adaoBuilder.New('%s')"%(self._objname, self._name))
        if self._content is not None:
            for command in self._content:
                self._append(*command)
    def _append(self, __command=None, __keys=None, __local=None, __pre=None, __switchoff=False):
        "Transformation d'une commande individuelle en un enregistrement"
        if __command is not None and __keys is not None and __local is not None:
            if "Concept" in __keys:
                logging.debug("TUI Order processed: %s"%(__local["Concept"],))
            __text  = ""
            if __pre is not None:
                __text += "%s = "%__pre
            __text += "%s.%s( "%(self._objname,str(__command))
            if "self" in __keys: __keys.remove("self")
            if __command not in ("set","get") and "Concept" in __keys: __keys.remove("Concept")
            for k in __keys:
                if k not in __local: continue
                __v = __local[k]
                if __v is None: continue
                if   k == "Checked"              and not __v: continue
                if   k == "Stored"               and not __v: continue
                if   k == "ColMajor"             and not __v: continue
                if   k == "InputFunctionAsMulti" and not __v: continue
                if   k == "nextStep"             and not __v: continue
                if   k == "PerformanceProfile"   and     __v: continue
                if   k == "noDetails":                        continue
                if isinstance(__v,Persistence.Persistence): __v = __v.values()
                if callable(__v): __text = self._missing%__v.__name__+__text
                if isinstance(__v,dict):
                    for val in __v.values():
                        if callable(val): __text = self._missing%val.__name__+__text
                numpy.set_printoptions(precision=15,threshold=1000000,linewidth=1000*15)
                __text += "%s=%s, "%(k,repr(__v))
                numpy.set_printoptions(precision=8,threshold=1000,linewidth=75)
            __text = __text.rstrip(", ")
            __text += " )"
            self._addLine(__text)
    def _extract(self, __multilines="", __object=None):
        "Transformation d'enregistrement(s) en commande(s) individuelle(s)"
        __is_case = False
        __commands = []
        __multilines = __multilines.replace("\r\n","\n")
        for line in __multilines.split("\n"):
            if "adaoBuilder.New" in line and "=" in line:
                self._objname = line.split("=")[0].strip()
                __is_case = True
                logging.debug("TUI Extracting commands of '%s' object..."%(self._objname,))
            if not __is_case:
                continue
            else:
                if self._objname+".set" in line:
                    __commands.append( line.replace(self._objname+".","",1) )
                    logging.debug("TUI Extracted command: %s"%(__commands[-1],))
        return __commands

class _COMViewer(GenericCaseViewer):
    """
    Établissement des commandes d'un cas COMM (Eficas Native Format/Cas<-COM)
    """
    def __init__(self, __name="", __objname="case", __content=None, __object=None):
        "Initialisation et enregistrement de l'entete"
        GenericCaseViewer.__init__(self, __name, __objname, __content, __object)
        self._observerIndex = 0
        self._addLine("# -*- coding: utf-8 -*-")
        self._addLine("#\n# Python script using ADAO COMM\n#")
        self._addLine("from numpy import array, matrix")
        self._addLine("#")
        self._addLine("%s = {}"%__objname)
        if self._content is not None:
            for command in self._content:
                self._append(*command)
    def _extract(self, __multilines=None, __object=None):
        "Transformation d'enregistrement(s) en commande(s) individuelle(s)"
        __suppparameters = {}
        if __multilines is not None:
            if 'adaoBuilder' in __multilines:
                raise ValueError("Impossible to load given content as an ADAO COMM one (Hint: it's perhaps not a COMM input, but a TUI one).")
            if "ASSIMILATION_STUDY" in __multilines:
                __suppparameters.update({'StudyType':"ASSIMILATION_STUDY"})
                __multilines = __multilines.replace("ASSIMILATION_STUDY","dict")
            elif "OPTIMIZATION_STUDY" in __multilines:
                __suppparameters.update({'StudyType':"ASSIMILATION_STUDY"})
                __multilines = __multilines.replace("OPTIMIZATION_STUDY",    "dict")
            elif "REDUCTION_STUDY" in __multilines:
                __suppparameters.update({'StudyType':"ASSIMILATION_STUDY"})
                __multilines = __multilines.replace("REDUCTION_STUDY",    "dict")
            elif "CHECKING_STUDY" in __multilines:
                __suppparameters.update({'StudyType':"CHECKING_STUDY"})
                __multilines = __multilines.replace("CHECKING_STUDY",    "dict")
            else:
                __multilines = __multilines.replace("ASSIMILATION_STUDY","dict")
            #
            __multilines = __multilines.replace("_F(",               "dict(")
            __multilines = __multilines.replace(",),);",             ",),)")
        __fulllines = ""
        for line in __multilines.split("\n"):
            if len(line) < 1: continue
            __fulllines += line + "\n"
        __multilines = __fulllines
        self._objname = "case"
        self._objdata = None
        exec("self._objdata = "+__multilines)
        #
        if self._objdata is None or not(type(self._objdata) is dict) or not('AlgorithmParameters' in self._objdata):
            raise ValueError("Impossible to load given content as an ADAO COMM one (no dictionnary or no 'AlgorithmParameters' key found).")
        # ----------------------------------------------------------------------
        logging.debug("COMM Extracting commands of '%s' object..."%(self._objname,))
        __commands = []
        __UserPostAnalysis = ""
        for k,r in self._objdata.items():
            __command = k
            logging.debug("COMM Extracted command: %s:%s"%(k, r))
            if   __command == "StudyName" and len(str(r))>0:
                __commands.append( "set( Concept='Name', String='%s')"%(str(r),) )
            elif   __command == "StudyRepertory":
                __commands.append( "set( Concept='Directory', String='%s')"%(str(r),) )
            elif   __command == "Debug" and str(r) == "0":
                __commands.append( "set( Concept='NoDebug' )" )
            elif   __command == "Debug" and str(r) == "1":
                __commands.append( "set( Concept='Debug' )" )
            elif   __command == "ExecuteInContainer":
                __suppparameters.update({'ExecuteInContainer':r})
            #
            elif __command == "UserPostAnalysis" and type(r) is dict:
                if 'STRING' in r:
                    __UserPostAnalysis = r['STRING'].replace("ADD",str(self._objname))
                    __commands.append( "set( Concept='UserPostAnalysis', String=\"\"\"%s\"\"\" )"%(__UserPostAnalysis,) )
                elif 'SCRIPT_FILE' in r and os.path.exists(r['SCRIPT_FILE']):
                    __UserPostAnalysis = open(r['SCRIPT_FILE'],'r').read()
                    __commands.append( "set( Concept='UserPostAnalysis', Script='%s' )"%(r['SCRIPT_FILE'],) )
                elif 'Template' in r and 'ValueTemplate' not in r:
                    # AnalysisPrinter...
                    if r['Template'] not in Templates.UserPostAnalysisTemplates:
                        raise ValueError("User post-analysis template \"%s\" does not exist."%(r['Template'],))
                    else:
                        __UserPostAnalysis = Templates.UserPostAnalysisTemplates[r['Template']]
                    __commands.append( "set( Concept='UserPostAnalysis', Template='%s' )"%(r['Template'],) )
                elif 'Template' in r and 'ValueTemplate' in r:
                    # Le template ayant pu être modifié, donc on ne prend que le ValueTemplate...
                    __UserPostAnalysis = r['ValueTemplate']
                    __commands.append( "set( Concept='UserPostAnalysis', String=\"\"\"%s\"\"\" )"%(__UserPostAnalysis,) )
                else:
                    __UserPostAnalysis = ""
            #
            elif __command == "AlgorithmParameters" and type(r) is dict and 'Algorithm' in r:
                if 'data' in r and r['Parameters'] == 'Dict':
                    __from = r['data']
                    if 'STRING' in __from:
                        __parameters = ", Parameters=%s"%(repr(eval(__from['STRING'])),)
                    elif 'SCRIPT_FILE' in __from and os.path.exists(__from['SCRIPT_FILE']):
                        __parameters = ", Script='%s'"%(__from['SCRIPT_FILE'],)
                else: # if 'Parameters' in r and r['Parameters'] == 'Defaults':
                    __Dict = copy.deepcopy(r)
                    __Dict.pop('Algorithm','')
                    __Dict.pop('Parameters','')
                    if 'SetSeed' in __Dict:__Dict['SetSeed'] = int(__Dict['SetSeed'])
                    if 'BoxBounds' in __Dict and type(__Dict['BoxBounds']) is str:
                        __Dict['BoxBounds'] = eval(__Dict['BoxBounds'])
                    if len(__Dict) > 0:
                        __parameters = ', Parameters=%s'%(repr(__Dict),)
                    else:
                        __parameters = ""
                __commands.append( "set( Concept='AlgorithmParameters', Algorithm='%s'%s )"%(r['Algorithm'],__parameters) )
            #
            elif __command == "Observers" and type(r) is dict and 'SELECTION' in r:
                if type(r['SELECTION']) is str:
                    __selection = (r['SELECTION'],)
                else:
                    __selection = tuple(r['SELECTION'])
                for sk in __selection:
                    __idata = r['%s_data'%sk]
                    if __idata['NodeType'] == 'Template' and 'Template' in __idata:
                        __template = __idata['Template']
                        if 'Info' in __idata:
                            __info = ", Info=\"\"\"%s\"\"\""%(__idata['Info'],)
                        else:
                            __info = ""
                        __commands.append( "set( Concept='Observer', Variable='%s', Template=\"\"\"%s\"\"\"%s )"%(sk,__template,__info) )
                    if __idata['NodeType'] == 'String' and 'Value' in __idata:
                        __value =__idata['Value']
                        __commands.append( "set( Concept='Observer', Variable='%s', String=\"\"\"%s\"\"\" )"%(sk,__value) )
            #
            # Background, ObservationError, ObservationOperator...
            elif type(r) is dict:
                __argumentsList = []
                if 'Stored' in r and bool(r['Stored']):
                    __argumentsList.append(['Stored',True])
                if 'INPUT_TYPE' in r and 'data' in r:
                    # Vector, Matrix, ScalarSparseMatrix, DiagonalSparseMatrix, Function
                    __itype = r['INPUT_TYPE']
                    __idata = r['data']
                    if 'FROM' in __idata:
                        # String, Script, DataFile, Template, ScriptWithOneFunction, ScriptWithFunctions
                        __ifrom = __idata['FROM']
                        __idata.pop('FROM','')
                        if __ifrom == 'String' or __ifrom == 'Template':
                            __argumentsList.append([__itype,__idata['STRING']])
                        if __ifrom == 'Script':
                            __argumentsList.append([__itype,True])
                            __argumentsList.append(['Script',__idata['SCRIPT_FILE']])
                        if __ifrom == 'DataFile':
                            __argumentsList.append([__itype,True])
                            __argumentsList.append(['DataFile',__idata['DATA_FILE']])
                        if __ifrom == 'ScriptWithOneFunction':
                            __argumentsList.append(['OneFunction',True])
                            __argumentsList.append(['Script',__idata.pop('SCRIPTWITHONEFUNCTION_FILE')])
                            if len(__idata)>0:
                                __argumentsList.append(['Parameters',__idata])
                        if __ifrom == 'ScriptWithFunctions':
                            __argumentsList.append(['ThreeFunctions',True])
                            __argumentsList.append(['Script',__idata.pop('SCRIPTWITHFUNCTIONS_FILE')])
                            if len(__idata)>0:
                                __argumentsList.append(['Parameters',__idata])
                __arguments = ["%s = %s"%(k,repr(v)) for k,v in __argumentsList]
                __commands.append( "set( Concept='%s', %s )"%(__command, ", ".join(__arguments)))
        #
        __commands.append( "set( Concept='%s', Parameters=%s )"%('SupplementaryParameters', repr(__suppparameters)))
        #
        # ----------------------------------------------------------------------
        __commands.sort() # Pour commencer par 'AlgorithmParameters'
        __commands.append(__UserPostAnalysis)
        return __commands

class _SCDViewer(GenericCaseViewer):
    """
    Établissement des commandes d'un cas SCD (Study Config Dictionary/Cas->SCD)

    Remarque : le fichier généré est différent de celui obtenu par EFICAS
    """
    def __init__(self, __name="", __objname="case", __content=None, __object=None):
        "Initialisation et enregistrement de l'entête"
        GenericCaseViewer.__init__(self, __name, __objname, __content, __object)
        #
        if __content is not None:
            for command in __content:
                if command[0] == "set": __command = command[2]["Concept"]
                else:                   __command = command[0].replace("set", "", 1)
                if __command == 'Name':
                    self._name = command[2]["String"]
        #
        self.__DebugCommandNotSet = True
        self.__ObserverCommandNotSet = True
        self.__UserPostAnalysisNotSet = True
        #
        self._addLine("# -*- coding: utf-8 -*-")
        self._addLine("#\n# Input for ADAO converter to SCD\n#")
        self._addLine("#")
        self._addLine("study_config = {}")
        self._addLine("study_config['Name'] = '%s'"%self._name)
        self._addLine("#")
        self._addLine("inputvariables_config = {}")
        self._addLine("inputvariables_config['Order'] =['adao_default']")
        self._addLine("inputvariables_config['adao_default'] = -1")
        self._addLine("study_config['InputVariables'] = inputvariables_config")
        self._addLine("#")
        self._addLine("outputvariables_config = {}")
        self._addLine("outputvariables_config['Order'] = ['adao_default']")
        self._addLine("outputvariables_config['adao_default'] = -1")
        self._addLine("study_config['OutputVariables'] = outputvariables_config")
        if __content is not None:
            for command in __content:
                self._append(*command)
    def _append(self, __command=None, __keys=None, __local=None, __pre=None, __switchoff=False):
        "Transformation d'une commande individuelle en un enregistrement"
        if __command == "set": __command = __local["Concept"]
        else:                  __command = __command.replace("set", "", 1)
        logging.debug("SCD Order processed: %s"%(__command))
        #
        __text  = None
        if __command in (None, 'execute', 'executePythonScheme', 'executeYACSScheme', 'get', 'Name'):
            return
        elif __command in ['Directory',]:
            __text  = "#\nstudy_config['Repertory'] = %s"%(repr(__local['String']))
        elif __command in ['Debug', 'setDebug']:
            __text  = "#\nstudy_config['Debug'] = '1'"
            self.__DebugCommandNotSet = False
        elif __command in ['NoDebug', 'setNoDebug']:
            __text  = "#\nstudy_config['Debug'] = '0'"
            self.__DebugCommandNotSet = False
        elif __command in ['Observer', 'setObserver']:
            if self.__ObserverCommandNotSet:
                self._addLine("observers = {}")
                self._addLine("study_config['Observers'] = observers")
                self.__ObserverCommandNotSet = False
            __obs   = __local['Variable']
            self._numobservers += 1
            __text  = "#\n"
            __text += "observers['%s'] = {}\n"%__obs
            if __local['String'] is not None:
                __text += "observers['%s']['nodetype'] = '%s'\n"%(__obs, 'String')
                __text += "observers['%s']['String'] = \"\"\"%s\"\"\"\n"%(__obs, __local['String'])
            if __local['Script'] is not None:
                __text += "observers['%s']['nodetype'] = '%s'\n"%(__obs, 'Script')
                __text += "observers['%s']['Script'] = \"%s\"\n"%(__obs, __local['Script'])
            if __local['Template'] is not None and __local['Template'] in Templates.ObserverTemplates:
                __text += "observers['%s']['nodetype'] = '%s'\n"%(__obs, 'String')
                __text += "observers['%s']['String'] = \"\"\"%s\"\"\"\n"%(__obs, Templates.ObserverTemplates[__local['Template']])
            if __local['Info'] is not None:
                __text += "observers['%s']['info'] = \"\"\"%s\"\"\"\n"%(__obs, __local['Info'])
            else:
                __text += "observers['%s']['info'] = \"\"\"%s\"\"\"\n"%(__obs, __obs)
            __text += "observers['%s']['number'] = %s"%(__obs, self._numobservers)
        elif __command in ['UserPostAnalysis', 'setUserPostAnalysis']:
            __text  = "#\n"
            __text += "Analysis_config = {}\n"
            if __local['String'] is not None:
                __text += "Analysis_config['From'] = 'String'\n"
                __text += "Analysis_config['Data'] = \"\"\"%s\"\"\"\n"%(__local['String'],)
            if __local['Script'] is not None:
                __text += "Analysis_config['From'] = 'Script'\n"
                __text += "Analysis_config['Data'] = \"\"\"%s\"\"\"\n"%(__local['Script'],)
            if __local['Template'] is not None and __local['Template'] in Templates.UserPostAnalysisTemplates:
                __text += "Analysis_config['From'] = 'String'\n"
                __text += "Analysis_config['Data'] = \"\"\"%s\"\"\"\n"%(Templates.UserPostAnalysisTemplates[__local['Template']],)
            __text += "study_config['UserPostAnalysis'] = Analysis_config"
            self.__UserPostAnalysisNotSet = False
        elif __local is not None: # __keys is not None and
            numpy.set_printoptions(precision=15,threshold=1000000,linewidth=1000*15)
            __text  = "#\n"
            __text += "%s_config = {}\n"%__command
            __local.pop('self','')
            __to_be_removed = []
            __vectorIsDataFile = False
            __vectorIsScript = False
            for __k,__v in __local.items():
                if __v is None: __to_be_removed.append(__k)
            for __k in __to_be_removed:
                __local.pop(__k)
            for __k,__v in __local.items():
                if __k == "Concept": continue
                if __k in ['ScalarSparseMatrix','DiagonalSparseMatrix','Matrix','OneFunction','ThreeFunctions'] \
                    and 'Script' in __local and __local['Script'] is not None: continue
                if __k in ['Vector','VectorSerie'] \
                    and 'DataFile' in __local and __local['DataFile'] is not None: continue
                if __k == 'Parameters' and not (__command in ['AlgorithmParameters','SupplementaryParameters']): continue
                if __k == 'Algorithm':
                    __text += "study_config['Algorithm'] = %s\n"%(repr(__v))
                elif __k == 'DataFile':
                    __k = 'Vector'
                    __f = 'DataFile'
                    __v = "'"+repr(__v)+"'"
                    for __lk in ['Vector','VectorSerie']:
                        if __lk in __local and __local[__lk]: __k = __lk
                    __text += "%s_config['Type'] = '%s'\n"%(__command,__k)
                    __text += "%s_config['From'] = '%s'\n"%(__command,__f)
                    __text += "%s_config['Data'] = %s\n"%(__command,__v)
                    __text = __text.replace("''","'")
                    __vectorIsDataFile = True
                elif __k == 'Script':
                    __k = 'Vector'
                    __f = 'Script'
                    __v = "'"+repr(__v)+"'"
                    for __lk in ['ScalarSparseMatrix','DiagonalSparseMatrix','Matrix']:
                        if __lk in __local and __local[__lk]: __k = __lk
                    if __command == "AlgorithmParameters": __k = "Dict"
                    if 'OneFunction' in __local and __local['OneFunction']:
                        __text += "%s_ScriptWithOneFunction = {}\n"%(__command,)
                        __text += "%s_ScriptWithOneFunction['Function'] = ['Direct', 'Tangent', 'Adjoint']\n"%(__command,)
                        __text += "%s_ScriptWithOneFunction['Script'] = {}\n"%(__command,)
                        __text += "%s_ScriptWithOneFunction['Script']['Direct'] = %s\n"%(__command,__v)
                        __text += "%s_ScriptWithOneFunction['Script']['Tangent'] = %s\n"%(__command,__v)
                        __text += "%s_ScriptWithOneFunction['Script']['Adjoint'] = %s\n"%(__command,__v)
                        __text += "%s_ScriptWithOneFunction['DifferentialIncrement'] = 1e-06\n"%(__command,)
                        __text += "%s_ScriptWithOneFunction['CenteredFiniteDifference'] = 0\n"%(__command,)
                        __k = 'Function'
                        __f = 'ScriptWithOneFunction'
                        __v = '%s_ScriptWithOneFunction'%(__command,)
                    if 'ThreeFunctions' in __local and __local['ThreeFunctions']:
                        __text += "%s_ScriptWithFunctions = {}\n"%(__command,)
                        __text += "%s_ScriptWithFunctions['Function'] = ['Direct', 'Tangent', 'Adjoint']\n"%(__command,)
                        __text += "%s_ScriptWithFunctions['Script'] = {}\n"%(__command,)
                        __text += "%s_ScriptWithFunctions['Script']['Direct'] = %s\n"%(__command,__v)
                        __text += "%s_ScriptWithFunctions['Script']['Tangent'] = %s\n"%(__command,__v)
                        __text += "%s_ScriptWithFunctions['Script']['Adjoint'] = %s\n"%(__command,__v)
                        __k = 'Function'
                        __f = 'ScriptWithFunctions'
                        __v = '%s_ScriptWithFunctions'%(__command,)
                    __text += "%s_config['Type'] = '%s'\n"%(__command,__k)
                    __text += "%s_config['From'] = '%s'\n"%(__command,__f)
                    __text += "%s_config['Data'] = %s\n"%(__command,__v)
                    __text = __text.replace("''","'")
                    __vectorIsScript = True
                elif __k in ('Stored', 'Checked', 'ColMajor', 'InputFunctionAsMulti', 'nextStep'):
                    if bool(__v):
                        __text += "%s_config['%s'] = '%s'\n"%(__command,__k,int(bool(__v)))
                elif __k in ('PerformanceProfile', 'noDetails'):
                    if not bool(__v):
                        __text += "%s_config['%s'] = '%s'\n"%(__command,__k,int(bool(__v)))
                else:
                    if __k == 'Vector' and __vectorIsScript: continue
                    if __k == 'Vector' and __vectorIsDataFile: continue
                    if __k == 'Parameters': __k = "Dict"
                    if isinstance(__v,Persistence.Persistence): __v = __v.values()
                    if callable(__v): __text = self._missing%__v.__name__+__text
                    if isinstance(__v,dict):
                        for val in __v.values():
                            if callable(val): __text = self._missing%val.__name__+__text
                    __text += "%s_config['Type'] = '%s'\n"%(__command,__k)
                    __text += "%s_config['From'] = '%s'\n"%(__command,'String')
                    __text += "%s_config['Data'] = \"\"\"%s\"\"\"\n"%(__command,repr(__v))
            __text += "study_config['%s'] = %s_config"%(__command,__command)
            numpy.set_printoptions(precision=8,threshold=1000,linewidth=75)
            if __switchoff:
                self._switchoff = True
        if __text is not None: self._addLine(__text)
        if not __switchoff:
            self._switchoff = False
    def _finalize(self, *__args):
        self.__loadVariablesByScript()
        if self.__DebugCommandNotSet:
            self._addLine("#\nstudy_config['Debug'] = '0'")
        if self.__UserPostAnalysisNotSet:
            self._addLine("#")
            self._addLine("Analysis_config = {}")
            self._addLine("Analysis_config['From'] = 'String'")
            self._addLine("Analysis_config['Data'] = \"\"\"import numpy")
            self._addLine("xa=ADD.get('Analysis')[-1]")
            self._addLine("print('Analysis:',xa)\"\"\"")
            self._addLine("study_config['UserPostAnalysis'] = Analysis_config")
    def __loadVariablesByScript(self):
        __ExecVariables = {} # Necessaire pour recuperer la variable
        exec("\n".join(self._lineSerie), __ExecVariables)
        study_config = __ExecVariables['study_config']
        # Pour Python 3 : self.__hasAlgorithm = bool(study_config['Algorithm'])
        if 'Algorithm' in study_config:
            self.__hasAlgorithm = True
        else:
            self.__hasAlgorithm = False
        if not self.__hasAlgorithm and \
                "AlgorithmParameters" in study_config and \
                isinstance(study_config['AlgorithmParameters'], dict) and \
                "From" in study_config['AlgorithmParameters'] and \
                "Data" in study_config['AlgorithmParameters'] and \
                study_config['AlgorithmParameters']['From'] == 'Script':
            __asScript = study_config['AlgorithmParameters']['Data']
            __var = ImportFromScript(__asScript).getvalue( "Algorithm" )
            __text = "#\nstudy_config['Algorithm'] = '%s'"%(__var,)
            self._addLine(__text)
        if self.__hasAlgorithm and \
                "AlgorithmParameters" in study_config and \
                isinstance(study_config['AlgorithmParameters'], dict) and \
                "From" not in study_config['AlgorithmParameters'] and \
                "Data" not in study_config['AlgorithmParameters']:
            __text  = "#\n"
            __text += "AlgorithmParameters_config['Type'] = 'Dict'\n"
            __text += "AlgorithmParameters_config['From'] = 'String'\n"
            __text += "AlgorithmParameters_config['Data'] = '{}'\n"
            self._addLine(__text)
        if 'SupplementaryParameters' in study_config and \
                isinstance(study_config['SupplementaryParameters'], dict) and \
                "From" in study_config['SupplementaryParameters'] and \
                study_config['SupplementaryParameters']["From"] == 'String' and \
                "Data" in study_config['SupplementaryParameters']:
            __dict = eval(study_config['SupplementaryParameters']["Data"])
            if 'ExecuteInContainer' in __dict:
                self._addLine("#\nstudy_config['ExecuteInContainer'] = '%s'"%__dict['ExecuteInContainer'])
            else:
                self._addLine("#\nstudy_config['ExecuteInContainer'] = 'No'")
            if 'StudyType' in __dict:
                self._addLine("#\nstudy_config['StudyType'] = '%s'"%__dict['StudyType'])
            if 'StudyType' in __dict and __dict['StudyType'] != "ASSIMILATION_STUDY":
                self.__UserPostAnalysisNotSet = False
        del study_config

class _YACSViewer(GenericCaseViewer):
    """
    Etablissement des commandes d'un cas YACS (Cas->SCD->YACS)
    """
    def __init__(self, __name="", __objname="case", __content=None, __object=None):
        "Initialisation et enregistrement de l'entete"
        GenericCaseViewer.__init__(self, __name, __objname, __content, __object)
        self.__internalSCD = _SCDViewer(__name, __objname, __content, __object)
        self._append       = self.__internalSCD._append
    def dump(self, __filename=None, __upa=None):
        "Restitution normalisée des commandes"
        # -----
        if __filename is None:
            raise ValueError("A file name has to be given for YACS XML output.")
        else:
            __file    = os.path.abspath(__filename)
            if os.path.isfile(__file) or os.path.islink(__file):
                os.remove(__file)
        # -----
        if not PlatformInfo.has_salome or \
            not PlatformInfo.has_adao:
            raise ImportError(
                "Unable to get SALOME (%s) or ADAO (%s) environnement for YACS conversion.\n"%(PlatformInfo.has_salome,PlatformInfo.has_adao)+\
                "Please load the right SALOME environnement before trying to use it.")
        else:
            from daYacsSchemaCreator.run import create_schema_from_content
        # -----
        self.__internalSCD._finalize(__upa)
        __SCDdump = self.__internalSCD.dump()
        create_schema_from_content(__SCDdump, __file)
        # -----
        if not os.path.exists(__file):
            __msg  = "An error occured during the ADAO YACS Schema build for\n"
            __msg += "the target output file:\n"
            __msg += "  %s\n"%__file
            __msg += "See errors details in your launching terminal log.\n"
            raise ValueError(__msg)
        # -----
        __fid = open(__file,"r")
        __text = __fid.read()
        __fid.close()
        return __text

# ==============================================================================
class _ReportViewer(GenericCaseViewer):
    """
    Partie commune de restitution simple
    """
    def __init__(self, __name="", __objname="case", __content=None, __object=None):
        "Initialisation et enregistrement de l'entete"
        GenericCaseViewer.__init__(self, __name, __objname, __content, __object)
        self._r = Reporting.ReportStorage()
        self._r.clear()
        if self._name == "":
            self._r.append("ADAO Study report", "title")
        else:
            self._r.append(str(self._name), "title")
        if self._content is not None:
            for command in self._content:
                self._append(*command)
    def _append(self, __command=None, __keys=None, __local=None, __pre=None, __switchoff=False):
        "Transformation d'une commande individuelle en un enregistrement"
        if __command is not None and __keys is not None and __local is not None:
            if __command in ("set","get") and "Concept" in __keys: __command = __local["Concept"]
            __text  = ""
            __text += "<i>%s</i> command has been set"%str(__command.replace("set",""))
            __ktext = ""
            for k in __keys:
                if k not in __local: continue
                __v = __local[k]
                if __v is None: continue
                if   k == "Checked"              and not __v: continue
                if   k == "Stored"               and not __v: continue
                if   k == "ColMajor"             and not __v: continue
                if   k == "InputFunctionAsMulti" and not __v: continue
                if   k == "nextStep"             and not __v: continue
                if   k == "PerformanceProfile"   and     __v: continue
                if   k == "noDetails":                        continue
                if   k == "Concept":                          continue
                if   k == "self":                             continue
                if isinstance(__v,Persistence.Persistence): __v = __v.values()
                numpy.set_printoptions(precision=15,threshold=1000000,linewidth=1000*15)
                __ktext += "\n        %s = %s,"%(k,repr(__v))
                numpy.set_printoptions(precision=8,threshold=1000,linewidth=75)
            if len(__ktext) > 0:
                __text += " with values:" + __ktext
            __text = __text.rstrip(", ")
            self._r.append(__text, "uli")
    def _finalize(self, __upa=None):
        "Enregistrement du final"
        raise NotImplementedError()

class _SimpleReportInRstViewer(_ReportViewer):
    """
    Restitution simple en RST
    """
    def _finalize(self, __upa=None):
        self._lineSerie.append(Reporting.ReportViewInRst(self._r).__str__())

class _SimpleReportInHtmlViewer(_ReportViewer):
    """
    Restitution simple en HTML
    """
    def _finalize(self, __upa=None):
        self._lineSerie.append(Reporting.ReportViewInHtml(self._r).__str__())

class _SimpleReportInPlainTxtViewer(_ReportViewer):
    """
    Restitution simple en TXT
    """
    def _finalize(self, __upa=None):
        self._lineSerie.append(Reporting.ReportViewInPlainTxt(self._r).__str__())

# ==============================================================================
class ImportFromScript(object):
    """
    Obtention d'une variable nommee depuis un fichier script importé
    """
    __slots__ = ("__basename", "__filenspace", "__filestring")
    def __init__(self, __filename=None):
        "Verifie l'existence et importe le script"
        if __filename is None:
            raise ValueError("The name of the file, containing the variable to be read, has to be specified.")
        if not os.path.isfile(__filename):
            raise ValueError(
                "The file containing the variable to be imported doesn't seem to"+\
                " exist. Please check the file. The given file name is:\n  \"%s\""%str(__filename))
        if os.path.dirname(__filename) != '':
            sys.path.insert(0, os.path.dirname(__filename))
            __basename = os.path.basename(__filename).rstrip(".py")
        else:
            __basename = __filename.rstrip(".py")
        PlatformInfo.checkFileNameImportability( __basename+".py" )
        self.__basename = __basename
        try:
            self.__filenspace = __import__(__basename, globals(), locals(), [])
        except NameError:
            self.__filenspace = ""
        with open(__filename,'r') as fid:
            self.__filestring = fid.read()
    def getvalue(self, __varname=None, __synonym=None ):
        "Renvoie la variable demandee par son nom ou son synonyme"
        if __varname is None:
            raise ValueError("The name of the variable to be read has to be specified. Please check the content of the file and the syntax.")
        if not hasattr(self.__filenspace, __varname):
            if __synonym is None:
                raise ValueError(
                    "The imported script file \"%s\""%(str(self.__basename)+".py",)+\
                    " doesn't contain the mandatory variable \"%s\""%(__varname,)+\
                    " to be read. Please check the content of the file and the syntax.")
            elif not hasattr(self.__filenspace, __synonym):
                raise ValueError(
                    "The imported script file \"%s\""%(str(self.__basename)+".py",)+\
                    " doesn't contain the mandatory variable \"%s\""%(__synonym,)+\
                    " to be read. Please check the content of the file and the syntax.")
            else:
                return getattr(self.__filenspace, __synonym)
        else:
            return getattr(self.__filenspace, __varname)
    def getstring(self):
        "Renvoie le script complet"
        return self.__filestring

# ==============================================================================
class ImportDetector(object):
    """
    Détection des caractéristiques de fichiers ou objets en entrée
    """
    __slots__ = (
        "__url", "__usr", "__root", "__end")
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        return False
    #
    def __init__(self, __url, UserMime=""):
        if __url is None:
            raise ValueError("The name or url of the file object has to be specified.")
        if __url is bytes:
            self.__url = __url.decode()
        else:
            self.__url = str(__url)
        if UserMime is bytes:
            self.__usr = UserMime.decode().lower()
        else:
            self.__usr = str(UserMime).lower()
        (self.__root, self.__end) = os.path.splitext(self.__url)
        #
        mimetypes.add_type('application/numpy.npy', '.npy')
        mimetypes.add_type('application/numpy.npz', '.npz')
        mimetypes.add_type('application/dymola.sdf', '.sdf')
        if sys.platform.startswith("win"):
            mimetypes.add_type('text/plain', '.txt')
            mimetypes.add_type('text/csv', '.csv')
            mimetypes.add_type('text/tab-separated-values', '.tsv')
    #
    # File related tests
    # ------------------
    def is_local_file(self):
        if os.path.isfile(os.path.realpath(self.__url)):
            return True
        else:
            return False
    def is_not_local_file(self):
        return not self.is_local_file()
    def raise_error_if_not_local_file(self):
        if self.is_not_local_file():
            raise ValueError("The name or the url of the file object doesn't seem to exist. The given name is:\n  \"%s\""%str(self.__url))
        else:
            return False
    #
    # Directory related tests
    # -----------------------
    def is_local_dir(self):
        if os.path.isdir(self.__url):
            return True
        else:
            return False
    def is_not_local_dir(self):
        return not self.is_local_dir()
    def raise_error_if_not_local_dir(self):
        if self.is_not_local_dir():
            raise ValueError("The name or the url of the directory object doesn't seem to exist. The given name is:\n  \"%s\""%str(self.__url))
        else:
            return False
    #
    # Mime related functions
    # ------------------------
    def get_standard_mime(self):
        (__mtype, __encoding) = mimetypes.guess_type(self.__url, strict=False)
        return __mtype
    def get_user_mime(self):
        __fake = "fake."+self.__usr.lower()
        (__mtype, __encoding) = mimetypes.guess_type(__fake, strict=False)
        return __mtype
    def get_comprehensive_mime(self):
        if self.get_standard_mime() is not None:
            return self.get_standard_mime()
        elif self.get_user_mime() is not None:
            return self.get_user_mime()
        else:
            return None
    #
    # Name related functions
    # ----------------------
    def get_user_name(self):
        return self.__url
    def get_absolute_name(self):
        return os.path.abspath(os.path.realpath(self.__url))
    def get_extension(self):
        return self.__end

class ImportFromFile(object):
    """
    Obtention de variables disrétisées en 1D, définies par une ou des variables
    nommées, et sous la forme d'une série de points éventuellement indexés. La
    lecture d'un fichier au format spécifié (ou intuité) permet de charger ces
    fonctions depuis :
        - des fichiers textes en colonnes de type TXT, CSV, TSV...
        - des fichiers de données binaires NPY, NPZ, SDF...
    La lecture du fichier complet ne se fait que si nécessaire, pour assurer la
    performance tout en disposant de l'interprétation du contenu. Les fichiers
    textes doivent présenter en première ligne (hors commentaire ou ligne vide)
    les noms des variables de colonnes. Les commentaires commencent par un "#".
    """
    __slots__ = (
        "_filename", "_colnames", "_colindex", "_varsline", "_format",
        "_delimiter", "_skiprows", "__url", "__filestring", "__header",
        "__allowvoid", "__binaryformats", "__supportedformats")
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        return False
    #
    def __init__(self, Filename=None, ColNames=None, ColIndex=None, Format="Guess", AllowVoidNameList=True):
        """
        Verifie l'existence et les informations de définition du fichier. Les
        noms de colonnes ou de variables sont ignorées si le format ne permet
        pas de les indiquer.
        Arguments :
            - Filename : nom du fichier
            - ColNames : noms de la ou des colonnes/variables à lire
            - ColIndex : nom unique de la colonne/variable servant d'index
            - Format : format du fichier et/ou des données inclues
            - AllowVoidNameList : permet, si la liste de noms est vide, de
              prendre par défaut toutes les colonnes
        """
        self.__binaryformats =(
            "application/numpy.npy",
            "application/numpy.npz",
            "application/dymola.sdf",
            )
        self.__url = ImportDetector( Filename, Format)
        self.__url.raise_error_if_not_local_file()
        self._filename = self.__url.get_absolute_name()
        PlatformInfo.checkFileNameConformity( self._filename )
        #
        self._format = self.__url.get_comprehensive_mime()
        #
        self.__header, self._varsline, self._skiprows = self.__getentete()
        #
        if self._format == "text/csv" or Format.upper() == "CSV":
            self._format = "text/csv"
            self.__filestring = "".join(self.__header)
            if self.__filestring.count(",") > 1:
                self._delimiter = ","
            elif self.__filestring.count(";") > 1:
                self._delimiter = ";"
            else:
                self._delimiter = None
        elif self._format == "text/tab-separated-values" or Format.upper() == "TSV":
            self._format = "text/tab-separated-values"
            self._delimiter = "\t"
        else:
            self._delimiter = None
        #
        if ColNames is not None: self._colnames = tuple(ColNames)
        else:                    self._colnames = None
        #
        if ColIndex is not None: self._colindex = str(ColIndex)
        else:                    self._colindex = None
        #
        self.__allowvoid = bool(AllowVoidNameList)

    def __getentete(self, __nblines = 3):
        "Lit l'entête du fichier pour trouver la définition des variables"
        # La première ligne non vide non commentée est toujours considérée
        # porter les labels de colonne, donc pas des valeurs
        __header, __varsline, __skiprows = [], "", 1
        if self._format in self.__binaryformats:
            pass
        else:
            with open(self._filename,'r') as fid:
                __line = fid.readline().strip()
                while "#" in __line or len(__line) < 1:
                    __header.append(__line)
                    __skiprows += 1
                    __line = fid.readline().strip()
                __varsline = __line # Ligne de labels par convention
                for i in range(max(0,__nblines)):
                    __header.append(fid.readline())
        return (__header, __varsline, __skiprows)

    def __getindices(self, __colnames, __colindex, __delimiter=None ):
        "Indices de colonnes correspondants à l'index et aux variables"
        if __delimiter is None:
            __varserie = self._varsline.strip('#').strip().split()
        else:
            __varserie = self._varsline.strip('#').strip().split(str(__delimiter))
        #
        if __colnames is not None:
            __usecols = []
            __colnames = tuple(__colnames)
            for v in __colnames:
                for i, n in enumerate(__varserie):
                    if v == n: __usecols.append(i)
            __usecols = tuple(__usecols)
            if len(__usecols) == 0:
                if self.__allowvoid:
                    __usecols = None
                else:
                    raise ValueError("Can not found any column corresponding to the required names %s"%(__colnames,))
        else:
            __usecols = None
        #
        if __colindex is not None:
            __useindex = None
            __colindex = str(__colindex)
            for i, n in enumerate(__varserie):
                if __colindex == n: __useindex = i
        else:
            __useindex = None
        #
        return (__usecols, __useindex)

    def getsupported(self):
        self.__supportedformats = {}
        self.__supportedformats["text/plain"]                = True
        self.__supportedformats["text/csv"]                  = True
        self.__supportedformats["text/tab-separated-values"] = True
        self.__supportedformats["application/numpy.npy"]     = True
        self.__supportedformats["application/numpy.npz"]     = True
        self.__supportedformats["application/dymola.sdf"]    = PlatformInfo.has_sdf
        return self.__supportedformats

    def getvalue(self, ColNames=None, ColIndex=None ):
        "Renvoie la ou les variables demandées par la liste de leurs noms"
        # Uniquement si mise à jour
        if ColNames is not None: self._colnames = tuple(ColNames)
        if ColIndex is not None: self._colindex = str(ColIndex)
        #
        __index = None
        if self._format == "application/numpy.npy":
            __columns = numpy.load(self._filename)
        #
        elif self._format == "application/numpy.npz":
            __columns = None
            with numpy.load(self._filename) as __allcolumns:
                if self._colnames is None:
                    self._colnames = __allcolumns.files
                for nom in self._colnames: # Si une variable demandée n'existe pas
                    if nom not in __allcolumns.files:
                        self._colnames = tuple( __allcolumns.files )
                for nom in self._colnames:
                    if nom in __allcolumns.files:
                        if __columns is not None:
                            # Attention : toutes les variables doivent avoir la même taille
                            __columns = numpy.vstack((__columns, numpy.reshape(__allcolumns[nom], (1,-1))))
                        else:
                            # Première colonne
                            __columns = numpy.reshape(__allcolumns[nom], (1,-1))
                if self._colindex is not None and self._colindex in __allcolumns.files:
                    __index = numpy.array(numpy.reshape(__allcolumns[self._colindex], (1,-1)), dtype=bytes)
        elif self._format == "text/plain":
            __usecols, __useindex = self.__getindices(self._colnames, self._colindex)
            __columns = numpy.loadtxt(self._filename, usecols = __usecols, skiprows=self._skiprows)
            if __useindex is not None:
                __index = numpy.loadtxt(self._filename, dtype = bytes, usecols = (__useindex,), skiprows=self._skiprows)
            if __usecols is None: # Si une variable demandée n'existe pas
                self._colnames = None
        #
        elif self._format == "application/dymola.sdf" and PlatformInfo.has_sdf:
            import sdf
            __content = sdf.load(self._filename)
            __columns = None
            if self._colnames is None:
                self._colnames = [__content.datasets[i].name for i in range(len(__content.datasets))]
            for nom in self._colnames:
                if nom in __content:
                    if __columns is not None:
                        # Attention : toutes les variables doivent avoir la même taille
                        __columns = numpy.vstack((__columns, numpy.reshape(__content[nom].data, (1,-1))))
                    else:
                        # Première colonne
                        __columns = numpy.reshape(__content[nom].data, (1,-1))
            if self._colindex is not None and self._colindex in __content:
                __index = __content[self._colindex].data
        #
        elif self._format == "text/csv":
            __usecols, __useindex = self.__getindices(self._colnames, self._colindex, self._delimiter)
            __columns = numpy.loadtxt(self._filename, usecols = __usecols, delimiter = self._delimiter, skiprows=self._skiprows)
            if __useindex is not None:
                __index = numpy.loadtxt(self._filename, dtype = bytes, usecols = (__useindex,), delimiter = self._delimiter, skiprows=self._skiprows)
            if __usecols is None: # Si une variable demandée n'existe pas
                self._colnames = None
        #
        elif self._format == "text/tab-separated-values":
            __usecols, __useindex = self.__getindices(self._colnames, self._colindex, self._delimiter)
            __columns = numpy.loadtxt(self._filename, usecols = __usecols, delimiter = self._delimiter, skiprows=self._skiprows)
            if __useindex is not None:
                __index = numpy.loadtxt(self._filename, dtype = bytes, usecols = (__useindex,), delimiter = self._delimiter, skiprows=self._skiprows)
            if __usecols is None: # Si une variable demandée n'existe pas
                self._colnames = None
        else:
            raise ValueError("Unkown file format \"%s\" or no reader available"%self._format)
        if __columns is None: __columns = ()

        def toString(value):
            try:
                return value.decode()
            except ValueError:
                return value
        if __index is not None:
            __index = tuple([toString(v) for v in __index])
        #
        return (self._colnames, __columns, self._colindex, __index)

    def getstring(self):
        "Renvoie le fichier texte complet"
        if self._format in self.__binaryformats:
            return ""
        else:
            with open(self._filename,'r') as fid:
                return fid.read()

    def getformat(self):
        return self._format

class ImportScalarLinesFromFile(ImportFromFile):
    """
    Importation de fichier contenant des variables scalaires nommées. Le
    fichier comporte soit 2, soit 4 colonnes, obligatoirement nommées "Name",
    "Value", "Minimum", "Maximum" si les noms sont précisés. Sur chaque ligne
    est indiqué le nom, la valeur, et éventuelement deux bornes min et max (ou
    None si nécessaire pour une borne).

    Seule la méthode "getvalue" est changée.
    """
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        return False
    #
    def __init__(self, Filename=None, ColNames=None, ColIndex=None, Format="Guess"):
        ImportFromFile.__init__(self, Filename, ColNames, ColIndex, Format)
        if self._format not in ["text/plain", "text/csv", "text/tab-separated-values"]:
            raise ValueError("Unkown file format \"%s\""%self._format)
    #
    def getvalue(self, VarNames = None, HeaderNames=()):
        "Renvoie la ou les variables demandées par la liste de leurs noms"
        if VarNames is not None: __varnames = tuple(VarNames)
        else:                    __varnames = None
        #
        if "Name" in self._varsline and "Value" in self._varsline and "Minimum" in self._varsline and "Maximum" in self._varsline:
            __ftype = "NamValMinMax"
            __dtypes   = {'names'  : ('Name', 'Value', 'Minimum', 'Maximum'),
                          'formats': ('S128', 'g', 'g', 'g')}
            __usecols  = (0, 1, 2, 3)

            def __replaceNoneN( s ):
                if s.strip() == b'None': return numpy.NINF
                else:                    return s

            def __replaceNoneP( s ):
                if s.strip() == b'None': return numpy.PINF
                else:                    return s
            __converters = {2: __replaceNoneN, 3: __replaceNoneP}
        elif "Name" in self._varsline and "Value" in self._varsline and ("Minimum" not in self._varsline or "Maximum" not in self._varsline):
            __ftype = "NamVal"
            __dtypes   = {'names'  : ('Name', 'Value'),
                          'formats': ('S128', 'g')}
            __converters = None
            __usecols  = (0, 1)
        elif len(HeaderNames)>0 and numpy.all([kw in self._varsline for kw in HeaderNames]):
            __ftype = "NamLotOfVals"
            __dtypes   = {'names'  : HeaderNames,
                          'formats': tuple(['S128',]+['g']*(len(HeaderNames)-1))}
            __usecols  = tuple(range(len(HeaderNames)))

            def __replaceNone( s ):
                if s.strip() == b'None': return numpy.NAN
                else:                    return s
            __converters = dict()
            for i in range(1,len(HeaderNames)):
                __converters[i] = __replaceNone
        else:
            raise ValueError("Can not find names of columns for initial values. Wrong first line is:\n            \"%s\""%self._varsline)
        #
        if self._format == "text/plain":
            __content = numpy.loadtxt(
                self._filename,
                dtype      = __dtypes,
                usecols    = __usecols,
                skiprows   = self._skiprows,
                converters = __converters,
                )
        elif self._format in ["text/csv", "text/tab-separated-values"]:
            __content = numpy.loadtxt(
                self._filename,
                dtype      = __dtypes,
                usecols    = __usecols,
                skiprows   = self._skiprows,
                converters = __converters,
                delimiter  = self._delimiter,
                )
        else:
            raise ValueError("Unkown file format \"%s\""%self._format)
        #
        __names, __thevalue, __bounds = [], [], []
        for sub in __content:
            if len(__usecols) == 4:
                na, va, mi, ma = sub
                if numpy.isneginf(mi): mi = None # Réattribue les variables None
                elif numpy.isnan(mi):  mi = None # Réattribue les variables None
                if numpy.isposinf(ma): ma = None # Réattribue les variables None
                elif numpy.isnan(ma):  ma = None # Réattribue les variables None
            elif len(__usecols) == 2 and __ftype == "NamVal":
                na, va = sub
                mi, ma = None, None
            else:
                nsub = list(sub)
                na = sub[0]
                for i, v in enumerate(nsub[1:]):
                    if numpy.isnan(v): nsub[i+1] = None
                va = nsub[1:]
                mi, ma = None, None
            na = na.decode()
            if (__varnames is None or na in __varnames) and (na not in __names):
                # Ne stocke que la premiere occurence d'une variable
                __names.append(na)
                __thevalue.append(va)
                __bounds.append((mi,ma))
        #
        __names      = tuple(__names)
        __thevalue = numpy.array(__thevalue)
        __bounds     = tuple(__bounds)
        #
        return (__names, __thevalue, __bounds)

# ==============================================================================
class EficasGUI(object):
    """
    Lancement autonome de l'interface EFICAS/ADAO
    """
    def __init__(self, __addpath = None):
        # Chemin pour l'installation (ordre important)
        self.__msg = ""
        self.__path_settings_ok = False
        #----------------
        if "EFICAS_ROOT" in os.environ:
            __EFICAS_ROOT = os.environ["EFICAS_ROOT"]
            __path_ok = True
        elif "EFICAS_NOUVEAU_ROOT" in os.environ:
            __EFICAS_ROOT = os.environ["EFICAS_NOUVEAU_ROOT"]
            __path_ok = True
        else:
            self.__msg += "\nKeyError:\n"+\
                "  the required environment variable EFICAS_ROOT is unknown.\n"+\
                "  You have either to be in SALOME environment, or to set\n"+\
                "  this variable in your environment to the right path \"<...>\"\n"+\
                "  to find an installed EFICAS application. For example:\n"+\
                "      EFICAS_ROOT=\"<...>\" command\n"
            __path_ok = False
        try:
            import adao
            __path_ok = True and __path_ok
        except ImportError:
            self.__msg += "\nImportError:\n"+\
                "  the required ADAO library can not be found to be imported.\n"+\
                "  You have either to be in ADAO environment, or to be in SALOME\n"+\
                "  environment, or to set manually in your Python 3 environment the\n"+\
                "  right path \"<...>\" to find an installed ADAO application. For\n"+\
                "  example:\n"+\
                "      PYTHONPATH=\"<...>:${PYTHONPATH}\" command\n"
            __path_ok = False
        try:
            import PyQt5
            __path_ok = True and __path_ok
        except ImportError:
            self.__msg += "\nImportError:\n"+\
                "  the required PyQt5 library can not be found to be imported.\n"+\
                "  You have either to have a raisonable up-to-date Python 3\n"+\
                "  installation (less than 5 years), or to be in SALOME environment.\n"
            __path_ok = False
        #----------------
        if not __path_ok:
            self.__msg += "\nWarning:\n"+\
                "  It seems you have some troubles with your installation.\n"+\
                "  Be aware that some other errors may exist, that are not\n"+\
                "  explained as above, like some incomplete or obsolete\n"+\
                "  Python 3, or incomplete module installation.\n"+\
                "  \n"+\
                "  Please correct the above error(s) before launching the\n"+\
                "  standalone EFICAS/ADAO interface.\n"
            logging.debug("Some of the ADAO/EFICAS/QT5 paths have not been found")
            self.__path_settings_ok = False
        else:
            logging.debug("All the ADAO/EFICAS/QT5 paths have been found")
            self.__path_settings_ok = True
        #----------------
        if self.__path_settings_ok:
            sys.path.insert(0,__EFICAS_ROOT)
            sys.path.insert(0,os.path.join(adao.adao_py_dir,"daEficas"))
            if __addpath is not None and os.path.exists(os.path.abspath(__addpath)):
                sys.path.insert(0,os.path.abspath(__addpath))
            logging.debug("All the paths have been correctly set up")
        else:
            print(self.__msg)
            logging.debug("Errors in path settings have been found")

    def gui(self):
        if self.__path_settings_ok:
            logging.debug("Launching standalone EFICAS/ADAO interface...")
            from daEficas import prefs
            from InterfaceQT4 import eficas_go
            eficas_go.lanceEficas(code=prefs.code)
        else:
            logging.debug("Can not launch standalone EFICAS/ADAO interface for path errors.")


# ==============================================================================
if __name__ == "__main__":
    print('\n AUTODIAGNOSTIC\n')



#-*- coding: utf-8 -*-
#
# Copyright (C) 2008-2022 EDF R&D
#
# This file is part of SALOME ADAO module
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See http://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author: Jean-Philippe Argaud, jean-philippe.argaud@edf.fr, EDF R&D

import sys
import os
import traceback
import logging

from daYacsSchemaCreator.infos_daComposant import *

def check_args(args):

  logging.debug("Arguments are :" + str(args))
  if len(args) != 2:
    raise ValueError("\n\n Bad number of arguments: you have to provide two arguments (%d given)\n" % (len(args)))

def check_study(study_config):

  logging.debug("[check_env] study_config : " + str(study_config))

  # Check study_config
  if not isinstance(study_config, dict):
    raise ValueError("\n\n Study configuration is not a dictionnary!\n")

  # Name
  if "Name" not in study_config:
    raise ValueError("\n\n Cannot find Name in the study configuration!\n")

  # Algorithm
  if "Algorithm" not in study_config:
    raise ValueError("\n\n Cannot find Algorithm in the study configuration!\n")
  else:
    if not (study_config["Algorithm"] in AssimAlgos or study_config["Algorithm"] in CheckAlgos):
      raise ValueError("\n\n Algorithm provided is unknow : " + str(study_config["Algorithm"]) +
                         "\n You can choose between : " + str(AssimAlgos)+" "+str(CheckAlgos) + "\n")

  # Debug
  if "Debug" not in study_config:
    study_config["Debug"] = "0"

  # Repertory
  check_repertory = False
  repertory = ""
  if "Repertory" in study_config.keys():
    repertory = study_config["Repertory"]
    check_repertory = True
    if not os.path.isabs(repertory):
      raise ValueError("\n\n Study repertory should be an absolute path\n"+
                           " Repertory provided is %s\n" % repertory)

  # ExecuteInContainer
  if "ExecuteInContainer" not in study_config:
    study_config["ExecuteInContainer"] = "No"

  # Check if all the data is provided
  for key in AlgoDataRequirements[study_config["Algorithm"]]:
    if key not in study_config.keys():
      raise ValueError("\n\nCannot find " +  key + " in your study configuration !" +
                    "\n This key is mandatory into a study with " + study_config["Algorithm"] + " algorithm." +
                    "\n " + study_config["Algorithm"] + " requirements are " + str(AlgoDataRequirements[study_config["Algorithm"]]) + "\n")

  # Data
  for key in study_config.keys():
    if key in AssimData:
      check_data(key, study_config[key], check_repertory, repertory)

  # UserDataInit
  if "UserDataInit" in study_config.keys():
    check_data("UserDataInit", study_config["UserDataInit"], check_repertory, repertory)

  # Variables
  check_variables("InputVariables", study_config)
  check_variables("OutputVariables", study_config)

  # Analyse
  if "UserPostAnalysis" in study_config.keys():
    analysis_config = study_config["UserPostAnalysis"]
    if "From" not in analysis_config:
      raise ValueError("\n\n UserPostAnalysis found but From is not defined \n in the analysis configuration!\n")
    else:
      if analysis_config["From"] not in AnalysisFromList:
        raise ValueError("\n\n Analysis From defined in the study configuration does not have a correct type : " + str(analysis_config["From"])
                         + "\n You can have: " + str(AnalysisFromList) + "\n")
    if "Data" not in analysis_config:
      raise ValueError("\n\nAnalysis found but Data is not defined in the analysis configuration!\n")

    if analysis_config["From"] == "Script":
      check_file_name = analysis_config["Data"]
      if check_repertory and not os.path.exists(check_file_name):
        check_file_name = os.path.join(repertory, os.path.basename(analysis_config["Data"]))
      if not os.path.exists(check_file_name):
        raise ValueError("\n\n The script file cannot be found for UserPostAnalysis,\n please check its availability.\n"+
                             " The given user file is:\n %s\n" % check_file_name)

  # Check observers
  if "Observers" in study_config.keys():
    for obs_var in study_config["Observers"]:
      # Check du type
      if not isinstance(study_config["Observers"][obs_var], type({})):
        raise ValueError("\n\n An observer description has to be a Python dictionary\n"+
                             " Observer is %s\n" % obs_var)
      if "nodetype" not in study_config["Observers"][obs_var].keys():
        raise ValueError("\n\n An observer description must provide a nodetype\n"+
                             " Observer is %s\n" % obs_var)
      nodetype = study_config["Observers"][obs_var]["nodetype"]
      if not isinstance(study_config["Observers"][obs_var]["nodetype"], type("")):
        raise ValueError("\n\n An observer nodetype description must be a string\n"+
                             " Observer is %s\n" % obs_var)
      if nodetype != "String" and nodetype != "Script":
        raise ValueError("\n\n An observer nodetype must be equal to 'String' or 'Script'\n"+
                             " Observer is %s\n" % obs_var)
      if nodetype == "String":
        if "String" not in study_config["Observers"][obs_var].keys():
          raise ValueError("\n\n An observer with nodetype String must provide a String\n"+
                               " Observer is %s\n" % obs_var)
        if not isinstance(study_config["Observers"][obs_var]["String"], type("")):
          raise ValueError("\n\n An observer String description must be a string\n"+
                               " Observer is %s\n" % obs_var)
      if nodetype == "Script":
        if "Script" not in study_config["Observers"][obs_var].keys():
          raise ValueError("\n\n An observer with nodetype Script provide a Script\n"+
                               " Observer is %s\n" % obs_var)
        if not isinstance(study_config["Observers"][obs_var]["Script"], type("")):
          raise ValueError("\n\n An observer Script description must be a string\n"+
                               " Observer is %s\n" % obs_var)
      if "scheduler" in study_config["Observers"][obs_var].keys():
        if not isinstance(study_config["Observers"][obs_var]["scheduler"], type("")):
          raise ValueError("\n\n An observer scheduler description must be a string\n"+
                              " Observer is %s\n" % obs_var)

def check_variables(name, study_config):

  if name not in study_config.keys():
    raise ValueError("\n\n %s not found in your study configuration!\n" % name)

  variable_config = study_config[name]
  if "Order" not in variable_config.keys():
    raise ValueError("\n\n Order not found in the %s configuration!\n" % name)

  list_of_variables = variable_config["Order"]
  if not isinstance(list_of_variables, type([])):
    raise ValueError("\n\n Order sould be a list in the %s configuration!\n" % name)
  if len(list_of_variables) < 1:
    raise ValueError("\n\nOrder should contain one or more names in the %s configuration!\n" % name)

  for var in list_of_variables:
    if var not in variable_config.keys():
      raise ValueError("\n\n Variable %s not found in the %s configuration!\n" % name)
    value = variable_config[var]
    try:
      value = int(value)
    except:
      raise ValueError("\n\n Variable %s value cannot be converted in an integer \n in the %s configuration!\n" % name)

def check_data(data_name, data_config, repertory_check=False, repertory=""):

  logging.debug("[check_data] " + data_name)
  data_name_data = "Data"
  data_name_type = "Type"
  data_name_from = "From"

  if data_name_data not in data_config:
    raise ValueError("\n\n" + data_name +" found but " + data_name_data +" is not defined in the study configuration !\n")

  if data_name_type not in data_config:
    raise ValueError("\n\n" + data_name +" found but " + data_name_type  +" is not defined in the study configuration !\n")
  else:
    if data_config[data_name_type] not in AssimType[data_name]:
      raise ValueError("\n\n" + data_name_type + " of " + data_name + " defined in the study configuration does not have a correct type : " + str(data_config[data_name_type])
                    + "\n You can have : " + str(AssimType[data_name]) + "\n")
  if data_name_from not in data_config:
    raise ValueError("\n\n" + data_name + " found but " + data_name_from + " is not defined in the study configuration !\n")
  else:
    if data_config[data_name_from] not in FromNumpyList[data_config[data_name_type]]:
      raise ValueError("\n\n" + data_name_from + " of " + data_name + " defined in the study configuration does not have a correct value : " + str(data_config[data_name_from])
                    + "\n You can have : " + str(FromNumpyList[data_config[data_name_type]]) + "\n")

  # Check des fichiers
  from_type = data_config["From"]
  if from_type == "Script":
    check_file_name = data_config["Data"]
    if repertory_check and not os.path.exists(check_file_name):
      check_file_name = os.path.join(repertory, os.path.basename(data_config["Data"]))
    if not os.path.exists(check_file_name):
      raise ValueError("\n\n The script file cannot be found for the \"%s\" keyword, please \n check its availability. The given user file is:\n %s\n"%(from_type,check_file_name))
  elif (from_type == "FunctionDict" or from_type == "ScriptWithSwitch" or from_type == "ScriptWithFunctions" or from_type == "ScriptWithOneFunction"):
    TheData = data_config["Data"]
    for FunctionName in TheData["Function"]:
      check_file_name = TheData["Script"][FunctionName]
      if repertory_check and not os.path.exists(check_file_name):
        check_file_name = os.path.join(repertory, os.path.basename(TheData["Script"][FunctionName]))
      if not os.path.exists(check_file_name):
        raise ValueError("\n\n The script file cannot be found for the \"%s\" keyword, please \n check its availability. The given user file is:\n %s\n"%(from_type,check_file_name))
